"""
Wrapper for ebas command line programs.

"""

import logging
import argparse
import getpass
import textwrap
import datetime
import re
from nilutility.argparse_helper import NiluArgumentParser, \
    ParseStrings, ParseIntegers, \
    ExclusiveMultiswithAction, \
    parse_time_instant, parse_time_interval
from nilutility.commandline import NiluCommandline
from nilutility.datetime_helper import DatetimeInterval
from nilutility.datatypes import DataObject
import six
from six.moves import input
from .io.file import EbasIOFormatOption, EbasIOFlagOption, \
    EBAS_IOSTYLE_SINGLECOLUMN, EBAS_IOSTYLE_MULTICOLUMN, \
    EBAS_IOSTYLE_MULTICOLUMN_PRECIP, EBAS_IOSTYLE_KEEP
from .db_fileindex import IndexDb, OperationalError

class EbasCommandline(NiluCommandline):  # pylint: disable-msg=R0902
    # R0902: Too many instance attributes
    """
    Class for handling Ebas commandline arguments.

    EbasCommandline is a wrapper for all ebas commandline programs, which
    handles all commandline arguments and the calls a callback function to the
    main program.
    Example usage:
        ```
        EbasCommandline(
            ebas_read_example,
            custom_args=['CONFIG', 'LOGGING', 'TIME_CRIT', 'NASA_READ'],
            private_args=add_private_args,
            help_description='%(prog)s example for reading a NasaAmes datafile.',
            version=__version__).run()
        ```

    The calling program specifies a callback method `func` (`ebas_read_example`
    in the example) which is called by the wrapper after the commandline
    arguments have been processed. The program logic is implemented in `func`
    in the caller's domain.

    `custom_args` defines the groups of arguments from the default ebas aruments
    which are needed by the calling program. For example

    - CONFIG adds the configuration file functionality
    - LOGGING adds the logfile aruments
    - TIME_CRIT adds time range argument
    - NASA_READ adds all arguments dealing with input files
    - ... many more

    `private_args` is a callback method in the callers domain which can set up
    additional arguments needed for the specific program.

    The commandline module in ebas can handle different levels of default values
    for all commandline arguments. There is a set of system default values for
    all arguments in the code.  
    When using the custom_args value CONFIG, the commandline module looks for a
    config file (called ebas.cfg by default) for reading default values which
    overrule the system default.  
    Thus, the user can set her own defaults. There is also the possibility to
    have config files in different places. The highest precedence has the file
    which is specified explicitly with ***full path name*** in the argument
    `--cfgfile` in the commanline, next is a config file in the current working
    directory, then in the user's home directory and finally a config file in
    the directory, the called program is located.

    All default values (system default or overruled by a config file) can of
    course be again overruled by using an explicit commandline argument.

    Settings in the config file can be done either on top level (Section EBAS4)
    for all programs, or for each program individually, or for a single program.
    Examples:
        ```
        [EBAS4]
        loglevfile=info
        ```
        Would generate by default a log file containing INFO messages or more severe for ALL programs.

        ```
        [EBAS4]

        [ebas_read_example]
        loglevfile=info
        ```
        Would generate by default this log file only for ebas_read_example.py
    """
    SUDOERS = ['paul', 'annehj', 'ebas', 'yong', 'richard', 'mye']

    def __init__(self, func, custom_args=None,  # pylint: disable-msg=R0913
                 private_args=None, help_description=None,
                 version='#no-version-number',
                 version_date=datetime.datetime.now(),
                 progname=None, def_args=None,
                 read_masterdata=True):
        NiluCommandline.__init__(
            self, func, custom_args=custom_args, private_args=private_args,
            help_description=help_description, version=version,
            version_date=version_date, progname=progname, def_args=def_args,
            config_main_section='EBAS4')
        self.dbh = None
        self.read_masterdata = read_masterdata
        self.versiontext = textwrap.dedent('''\n
            EBAS %(prog)s {0}
            Copyright (C) 2012-{1} NILU - Norwegian Institut for Air Reserch
            '''.format(self.version, self.version_date.year))
        self.logger = logging.getLogger('EbasCmdLine')

    def run(self):
        """
        Processes the arguments, then calls the `func` callback method.  
        Always ends the program with a DB rollback.
        Programs shoiuld commit on success before exiting.
        Programs crashing, or just exiting without commit will by this rollback
        mechanism clean up open TR entries.
        Works with ctrl-c, sysexit and all kinds of exceptions.
        Works also with --help and --version.
        """
        try:
            NiluCommandline.run(self)
        finally:
            if self.dbh:
                self.dbh.rollback()

    def preparse(self):
        NiluCommandline.preparse(self)
        if 'DB' in self.custom_args or 'DBWRITE' in self.custom_args:
            self.preparse_db()

    def set_defaults(self):
        """
        Set defaults.

        Parameters:
            None
        Returns:
           None
        """
        NiluCommandline.set_defaults(self)
        if 'CONFIG' in self.custom_args:
            self.def_args['cfgfile'] = 'ebas.cfg'
        if 'DB' in self.custom_args or 'DBWRITE' in self.custom_args:
            self.def_args['dbHost'] = 'prod-ebas-pg'
            self.def_args['db'] = 'ebas'
            self.def_args['dbUser'] = getpass.getuser()
            self.def_args['dbPasswd'] = None

    def preparse_db(self):
        """
        Setup db connection.
        Partially parse commandline arguments relevant for db, then connect
        according to setings.
        Needs to be pre-parsed in order to use master data from the DB to
        validate other arguments.
        """
        from ebas.db4 import EbasDb, DbError
        preparser = NiluArgumentParser(add_help=False)
        preparser.add_db_args()
        preparser.set_known_defaults(**self.def_args)
        self.args, self.unparsed = preparser.parse_known_args(
            args=self.unparsed, namespace=self.args)

        if self.args.nodb:
            self.logger.info('DB connection disabled by argument --nodb')
            return
        if not self.args.db or not self.args.dbHost:
            self.logger.error(
                'no DB server or database name specified, no DB connection '
                'possible')
            return
        while not self.args.dbUser:
            self.args.dbUser = input(
                'user name for database {0}: '.format(self.args.db))
        #while self.args.dbPasswd is None or self.args.dbPasswd == '':
        #    self.args.dbPasswd = getpass.getpass(
        #        'password for user {0} at database {1}: '
        #        .format(self.args.dbUser, self.args.db))
        try:
            self.dbh = EbasDb(self.args.dbHost, self.args.dbUser,
                              self.args.dbPasswd, self.args.db,
                              port=self.args.dbPort,
                              readonly=not 'DBWRITE' in self.custom_args,
                              progname=self.progname, progvers=self.version)
        except DbError as exp:
            self.logger.error(exp)
            self.dbh = None
        if self.dbh and self.read_masterdata:
            from ebas.domain.masterdata import read_all_caches_from_db
            read_all_caches_from_db(self.dbh)

    def add_custom_args(self, parser):
        """
        Add arguments for specific customized goals. E.g. custom parser
        'DS_CRIT' activates all arguments used to select datasets.

        Parameters:
            parser    parser object
        Returns:
            None
        """
        NiluCommandline.add_custom_args(self, parser)
        if 'DBWRITE' in self.custom_args:
            parser.add_db_args(nowrite=True, nocommit=True, transcomment=True)
        elif 'DB' in self.custom_args:
            parser.add_db_args(nowrite=False, nocommit=False,
                               transcomment=False)
        if 'DS_CRIT' in self.custom_args:
            parser_ds_crit_group = parser.add_argument_group(
                'dataset selection criteria')
            option_type = option_bool('AuxiliaryOption',
                                      'include-auxiliary',
                                      'exclude-auxiliary')
            parser_ds_crit_group.add_argument(
                '--include-auxiliary', '--exclude-auxiliary',
                type=option_type,
                action=ExclusiveMultiswithAction,
                default='include-auxiliary',
                dest='auxiliary',
                help='include or exclude auxiliary datasets')
            option_type = option_bool('NonAuxiliaryOption',
                                      'include-nonauxiliary',
                                      'exclude-nonauxiliary')
            parser_ds_crit_group.add_argument(
                '--include-nonauxiliary', '--exclude-nonauxiliary',
                type=option_type,
                action=ExclusiveMultiswithAction,
                default='include-nonauxiliary',
                dest='nonauxiliary',
                help='include or exclude non auxiliary datasets')
            option_type = option_bool('NRTOption',
                                      'include-nrt',
                                      'exclude-nrt')
            parser_ds_crit_group.add_argument(
                '--include-nrt', '--exclude-nrt',
                type=option_type,
                action=ExclusiveMultiswithAction,
                default='include-nrt',
                dest='nrt',
                help='include or exclude NRT datasets')
            option_type = option_bool('NonNRTOption',
                                      'include-non-nrt',
                                      'exclude-non-nrt')
            parser_ds_crit_group.add_argument(
                '--include-non-nrt', '--exclude-non-nrt',
                type=option_type,
                action=ExclusiveMultiswithAction,
                default='include-non-nrt',
                dest='nonnrt',
                help='include or exclude non NRT datasets')
            parser_ds_crit_group.add_argument('--dw_id', type=int,
                                              help='download id from database')
            parser_ds_crit_group.add_argument(
                '--setkey', '-k ',
                type=ParseIntegers(allow_scalar=True, allow_list=True,
                                   allow_range=True),
                default=None,
                help='dataset setkey; int, list of int or min-max')
            parser_ds_crit_group.add_argument(
                '--su_id',
                type=ParseIntegers(allow_scalar=True, allow_list=True,
                                   allow_range=True),
                default=None,
                help='submission identifier; int, list of int or min-max')
            parser_ds_crit_group.add_argument(
                '--station', '-s ',
                type=ValidateArgument(self, ValidateArgument.val_stationcode),
                default=None,
                help='station code or station code list')
            parser_ds_crit_group.add_argument(
                '--project', '-p ',
                type=ValidateArgument(self, ValidateArgument.val_project),
                default=None,
                help='project or project list')
            parser_ds_crit_group.add_argument(
                '--exproject',
                type=ValidateArgument(self, ValidateArgument.val_project),
                default=None,
                help='exclude project or project list (experimental, probably '
                'removed again')
            parser_ds_crit_group.add_argument(
                '--instrument', '-i ',
                type=ValidateArgument(self, ValidateArgument.val_instrument),
                default=None,
                help='instrument type or instrument type list')
            parser_ds_crit_group.add_argument(
                '--component', '-c ',
                type=ValidateArgument(self, ValidateArgument.val_component),
                default=None,
                help='component name or component name list. Strict synonyms '
                'are allowed, lookup names are accepted case insensitive as '
                'long as this is not in conflicht with other names.')
            parser_ds_crit_group.add_argument(
                '--matrix', '-m ',
                type=ValidateArgument(self, ValidateArgument.val_matrix),
                default=None,
                help='matrix name or matrix name list')
            parser_ds_crit_group.add_argument(
                '--group', '-g ',
                type=ValidateArgument(self, ValidateArgument.val_group),
                default=None,
                help='parameter group or parameter group list:\n' + 'bla')
            parser_ds_crit_group.add_argument(
                '--fi_ref',
                type=ValidateArgument(self, ValidateArgument.val_fi_ref),
                default=None,
                help='field instrument reference, or field instrument '
                     'reference list')
            parser_ds_crit_group.add_argument(
                '--me_ref',
                type=ValidateArgument(self, ValidateArgument.val_me_ref),
                default=None,
                help='method reference or method reference list')
            parser_ds_crit_group.add_argument(
                '--resolution',
                type=ValidateArgument(self, ValidateArgument.val_resolution),
                default=None,
                help='time resolution or time resolution list')
            parser_ds_crit_group.add_argument(
                '--statistics',
                type=ValidateArgument(self, ValidateArgument.val_statistics),
                default=None,
                help='statistical dimension or list of statistical dimensions')
            parser_ds_crit_group.add_argument(
                '--datalevel',
                type=ValidateArgument(self, ValidateArgument.val_datalevel),
                default=None,
                help='data level')
            parser_ds_crit_group.add_argument(
                '--characteristics',
                type=ValidateArgument(
                    self, ValidateArgument.val_characteristics),
                default=None,
                help='characteristics criteria: characteristic_type=value[,...]')
            # TODO: HDS, HVM, DP, SU, DC criteria
        if 'TIME_CRIT' in self.custom_args:
            parser_ds_crit_group = parser.add_argument_group('time selection')
            parser_ds_crit_group.add_argument(
                '--time', '-t ', type=parse_time_interval,
                default=DatetimeInterval(None, None),
                help=textwrap.dedent('''
                time criteria: format FROM[-TO].
                FROM and TO each has the format YYYY[-MM[-DD[THH[:MM[:SS]]]]].

                If only one time is specified, the period which is defined by\
                the precision of the time format will be chosen:
                2011-01 --> [2011-01-01T00:00:00,2011-02-01T00:00:00[\
                (i.e. the period January 2011).
                2011-11-23T11 --> [11:00, 12:00[ on 23 Nov.

                If both FROM and TO are given, the interval is defined by the\
                start of the FROM period and the end of the TO period:
                2011-01-2012-01-22 -->\
                [2011-01-01T00:00:00, 2012-01-23T00:00:00[

                The time criteria works by overlapping the given criteria with\
                the measurement sample sequence.
                When time is specified as YYYY or YYYY-MM, a slightly different\
                approach is used: sample sequences that overlap only partly\
                with the first or last sample in the sequence are not included.\
                (This is to prevent non-intuitive overlaps when one e.g.\
                07:00-07:00 sample overlaps from the last month or year of data)
                With other words: when specifying only YYYY or YYYY-MM, a hit\
                is only considered, if at least 1 _full_ sample is within the\
                respective year or month.'''))
        if 'STATE_CRIT' in self.custom_args:
            parser_ts_crit_group = parser.add_argument_group('database state')
            parser_ts_crit_group.add_argument(
                '--state', type=parse_time_instant, default=None,
                help=textwrap.dedent('''
                    state criteria: format YYYY-MM-DDTHH:MM:SS.MMM
                    Defines the (historic) state of the database which should \
                    be shown.'''))
            parser_ts_crit_group.add_argument(
                '--diff', type=parse_time_instant, default=None,
                help=textwrap.dedent('''
                    diff criteria: format YYYY-MM-DDTHH:MM:SS.MMM
                    Only data changed between this timestamp and the requested \
                    state will be considered.'''))
        if 'US_ID' in self.custom_args:
            username = getpass.getuser()
            if username in self.__class__.SUDOERS:
                parser_us_group = parser.add_argument_group('on behalf of...')
                parser_us_group.add_argument(
                    '--us_id', type=int, default=None,
                    help=textwrap.dedent('''
                        data access will be done on behalf of the specified \
                        EBAS user ID (US_ID). Use 1 for anonymous access'''))
        if 'NASA_READ' in self.custom_args:
            parser_nasa_group = parser.add_argument_group(
                'nasa ames reading options')
            parser_nasa_group.add_argument(
                '--ignore',
                type=ParseStrings(
                    allow_scalar=True, allow_list=True, allowed_values=[
                        'numformat', 'dx0', 'rescode', 'sample-duration',
                        'identical-values', 'revdate', 'parameter',
                        'flag-consistency', 'value-check', 'template-check']),
                default=[],
                help='ignore certain errors while reading (numformat, dx0, '
                'rescode, identical-values, flag-consistency, value-check)\n\n'
                'numformat: ignore if number format does not match missing '
                'value code\n\n'
                'dx0: allow DX=0 with regular data (DX!=0 will always raise an '
                'error if not correct)\n\n'
                'rescode: ignore resolution code errors\n\n'
                'sample-duration: ignore sample duration code errors\n\n'
                'identical-values: ignore errors related to all values of one '
                'variable being equal (e.g. all data missing for one variable, '
                'all values 0.0, or all values any other constant value). '
                'Those cases usually indicate an ERROR in the data. However if '
                'this is investigated and those data should be read, use this '
                'flag to downgrade the ERROR severity to WARNING. For the '
                'opposite (those cases are definitely identified as ERROR '
                'should not be read while the rest of the file should be read, '
                'consider the --skip-variables parameter)\n\n'
                'parameter: ignore validity of parameters (component names or '
                'component/matrix combination.\n\n'
                'flag-consistency: ignore checks for flag validity, flag '
                'consistency and value/flag consistency\n'
                'Those issues will be downgraded to WARNING.\n'
                '!!! Please be aware that this only applies to the reading of '
                'input files. For example, ebas_insert will _still_ come up '
                'with error messages regarding flag consistency and illegal '
                'flags!!!\n\n'
                'value-check: ignore value checks (boundaries and spikes)\n'
                'Those issues will be downgraded to WARNING.\n'
                '!!! Please be aware that this only applies to the reading of '
                'input files. For example, ebas_insert will _still_ come up '
                'with error messages regarding boundary and spike checks!!!\n\n')
            parser_nasa_group.add_argument(
                '--fix',
                type=ParseStrings(
                    allow_scalar=True, allow_list=True, allowed_values=[
                        'flag-consistency', 'overlap_set_start', 'bc211']),
                default=[],
                help='fix certain errors while reading (flag consistency, '
                'sample overlap, flag away boundary checks).\n\n'
                'flag-consistency: fix problems with flag consistency (e.g. '
                'missing flag for valid value, flag 100 for missing data, ...'
                '\n\noverlap_set_start: Errors with overlapping sample times '
                'are fixed by adjusting the start time of the second sample'
                '\n\nbc211: in case of boundary check violations, add flag 211')
            parser_nasa_group.add_argument(
                '--condense-messages',
                type=ParseIntegers(allow_scalar=True, allow_list=False,
                                   allow_range=False),
                default=8,
                help='Threshold for condensation of multiple occurrences of '
                'the same error or warning message. 0 turns off condensation '
                '(each single message will be logged)')
            parser_nasa_group.add_argument(
                '--skip-variables',
                type=ParseIntegers(allow_scalar=True, allow_list=True,
                                   allow_range=False),
                default=[],
                help='skip the variable(s) referenced by the variable '
                'number(s) while reading the file.\n'
                'Specify a single variable number or a comma separated list of '
                'variable numbers. '
                'Variable numbers start with 1 for the variable succeeding '
                'end_time (same numbering as in all WARNING/ERROR messages)\n'
                'start_time and end_time can not be skipped, numflags may not '
                'be skipped')
            parser_nasa_group.add_argument(
                '--skip-data', action='store_true',
                help='do not read data from the file, only header information '
                'and metadata.')
            option_type = option_bool('UnitConvertOption',
                                      'unitconvert-read',
                                      'no-unitconvert-read')
            parser_nasa_group.add_argument(
                '--unitconvert-read', '--no-unitconvert-read',
                type=option_type,
                action=ExclusiveMultiswithAction,
                default='unitconvert-read',
                dest='unitconvert_read',
                help='No conversion of units when reading a file.')
        elif 'OTHER_READ' in self.custom_args:
            parser_file_group = parser.add_argument_group(
                'file reading options')
            parser_file_group.add_argument(
                '--ignore',
                type=ParseStrings(allow_scalar=True, allow_list=True,
                                  allowed_values=['identical-values',
                                                  'parameter',
                                                  'flag-consistency',
                                                  'value-check']),
                default=[],
                help='ignore certain errors while reading (identical-values, '
                'flag-consistency, value-check)\n\n'
                'identical-values: ignore errors related to all values of one '
                'variable being equal (e.g. all data missing for one variable, '
                'all values 0.0, or all values any other constant value). '
                'Those cases usually indicate an ERROR in the data. However if '
                'this is investigated and those data should be read, use this '
                'flag to downgrade the ERROR severity to WARNING. For the '
                'opposite (those cases are definitely identified as ERROR '
                'should not be read while the rest of the file should be read, '
                'consider the --skip-variables parameter)\n\n'
                'parameter: ignore validity of parameters (component names or '
                'component/matrix combination.\n\n'
                'flag-consistency: ignore checks for flag validity, flag '
                'consistency and value/flag consistency\n'
                'Those issues will be downgraded to WARNING.\n'
                '!!! Please be aware that this only applies to the reading of '
                'input files. For example, ebas_insert will _still_ come up '
                'with error messages regarding flag consistency and illegal '
                'flags!!!\n\n'
                'value-check: ignore value checks (boundaries and spikes)\n'
                'Those issues will be downgraded to WARNING.\n'
                '!!! Please be aware that this only applies to the reading of '
                'input files. For example, ebas_insert will _still_ come up '
                'with error messages regarding boundary and spike checks!!!\n\n')
            option_type = option_bool('UnitConvertOption',
                                      'unitconvert-read',
                                      'no-unitconvert-read')
            parser_file_group.add_argument(
                '--unitconvert-read', '--no-unitconvert-read',
                type=option_type,
                action=ExclusiveMultiswithAction,
                default='unitconvert-read',
                dest='unitconvert_read',
                help='No conversion of units when reading a file.')
        if 'FILE_OUTPUT' in self.custom_args or \
                'FILE_OUTPUT_KEEPCOLUMNS' in self.custom_args:
            parser_output_group = parser.add_argument_group(
                'file output options')
            parser_output_group.add_argument(
                '--format', type=EbasIOFormatOption,
                default=EbasIOFormatOption('NasaAmes'),
                help='file format for output: ({})'.format(
                    ', '.join(
                        [str(x) for x in EbasIOFormatOption.legal_options()])))
            if 'FILE_OUTPUT_KEEPCOLUMNS' in self.custom_args:
                option_type = option_int(
                    'MulticolumnOption', {
                        'multicolumn': EBAS_IOSTYLE_MULTICOLUMN,
                        'singlecolumn': EBAS_IOSTYLE_SINGLECOLUMN,
                        'multicolumn-precipitation':
                            EBAS_IOSTYLE_MULTICOLUMN_PRECIP,
                        'keep-columns': EBAS_IOSTYLE_KEEP,
                    })
                parser_output_group.add_argument(
                    '--multicolumn', '--singlecolumn',
                    '--multicolumn-precipitation', '--keep-columns',
                    type=option_type,
                    action=ExclusiveMultiswithAction,
                    default='keep-columns',
                    dest='style',
                    help='singlecolumn output or multicolumn output '
                         '(one or more variables per file\n'
                         '--multicolumn-precipitation: generally singlecolumn '
                         'output, but add precipitation_amount as additional '
                         'variables\n'
                         '--keep-columns: keep the columns as they are in the '
                         'source file')
            else:
                option_type = option_int(
                    'MulticolumnOption', {
                        'multicolumn': EBAS_IOSTYLE_MULTICOLUMN,
                        'singlecolumn': EBAS_IOSTYLE_SINGLECOLUMN,
                        'multicolumn-precipitation':
                            EBAS_IOSTYLE_MULTICOLUMN_PRECIP,
                    })
                parser_output_group.add_argument(
                    '--multicolumn', '--singlecolumn',
                    '--multicolumn-precipitation',
                    type=option_type,
                    action=ExclusiveMultiswithAction,
                    default='multicolumn',
                    dest='style',
                    help='singlecolumn output or multicolumn output '
                         '(one or more variables per file\n'
                         '--multicolumn-precipitation: generally singlecolumn '
                         'output, but add precipitation_amount as additional '
                         'variables')
            option_type = option_bool('LongTimeseriesOption',
                                      'long_timeseries', 'no-long_timeseries')
            parser_output_group.add_argument(
                '--long_timeseries', '--no-long_timeseries',
                type=option_type,
                action=ExclusiveMultiswithAction,
                default='no-long_timeseries',
                dest='long_timeseries',
                help='make long timeseries, even when metadata change. Only '
                     'supported for file formats which support changing '
                     'metadata (OPenDAP, NetCDF)')
            option_type = option_bool('CreateFilesOption',
                                      'createfiles', 'stdout')
            parser_output_group.add_argument(
                '--createfiles', '--stdout',
                type=option_type,
                action=ExclusiveMultiswithAction,
                default='createfiles',
                dest='createfiles',
                help='create files or output to stdout')
            option_type = option_bool('XmlWrapOption',
                                      'xmlwrap', 'no-xmlwrap')
            parser_output_group.add_argument(
                '--xmlwrap', '--no-xmlwrap',
                type=option_type,
                action=ExclusiveMultiswithAction,
                default='no-xmlwrap',
                dest='xmlwrap',
                help='wrap output in xml containers (only for --stdout)')
            option_type = option_bool('UnitConvertOption',
                                      'unitconvert-write',
                                      'no-unitconvert-write')
            parser_output_group.add_argument(
                '--unitconvert-write', '--no-unitconvert-write',
                type=option_type,
                action=ExclusiveMultiswithAction,
                default='unitconvert-write',
                dest='unitconvert_write',
                help='For some parameters, additional variables with different '
                     'units will be generated on export (e.g. VOCs or ozone in '
                     'nmol/mol, nitrate and sulphate in ug/m3, ...).\n\n'
                     'With the option --no-unitconvert-write, the additional '
                     'variables will not be exported. Only the variables with '
                     'the default units stored in EBAS will be exported.')
            parser_output_group.add_argument(
                '--destdir', action=ParseDestDirAction,
                help='set output directory for files')

            parser_output_group.add_argument(
                '--flags',
                choices=EbasIOFlagOption.legal_options(), action=None,
                default=EbasIOFlagOption('one-or-all'), type=EbasIOFlagOption,
                help=textwrap.dedent('''
                    flag columns style:

                    * one-or-all (default):
                    If all variables share the same sequence of flags throughout\
                    the whole file, use one flag column as last column.
                    Else, one flag column per variable is used.
                    This is the default behavior starting from EBAS 3.0.

                    * compress:
                    If multiple variables share the same sequence of flags\
                    throughout the whole file, one flag column after this group\
                    of variables is used.
                    This produces files as narrow as possible without losing any flag\
                    information.
                    This used to be the default behavior up to EBAS 2.2.

                    * all
                    All variables get a dedicated flag column.

                    * none
                    No flag columns are exported. Invalid or missing data are both\
                    reported as MISSING value. This should be used very carfully,\
                    as information is LOST on export!
                    Intended for non expert uses, as the easiest approach to process\
                    only valid data, without bothering about the EBAS flag system.
                    Note: Detection limit values (flag 781) are exported as\
                    value/2.0 (only in this case, when no flag information is \
                    extracted).

                    As a general rule, a flag column applies ALWAYS to all preceding\
                    variables after the previous flag column.'''))
            parser_output_group.add_argument(
                '--fileindex', type=parse_fileindex,
                help='file path and name for ebas file index database (sqlite3).'
                     'Prepend a plus sign (+) to the filename in order to add to an '
                     'existing database instead of creating a new one.')

    def get_custom_args(self, custom_parser):
        """
        Get custom parser arguments in a separate dictionary.

        Parameters:
            custom_parser    parser group name
        Returns:
            {}
        """
        ret = DataObject()
        translate = \
            {
                'DS_CRIT':
                {
                    'dw_id': 'DW_ID',
                    'su_id': 'SU_ID',
                    'setkey': 'DS_SETKEY',
                    'station': 'ST_STATION_CODE',
                    'project': 'PR_ACRONYM',
                    'exproject': '!PR_ACRONYM',
                    'instrument': 'FT_TYPE',
                    'component': 'CO_COMP_NAME',
                    'matrix': 'MA_MATRIX_NAME',
                    'group': 'PG_NAME',
                    'fi_ref': 'FI_REF',
                    'me_ref': 'ME_REF',
                    'resolution': 'DS_RESCODE',
                    'statistics': 'SC_STATISTIC_CODE',
                    'datalevel': 'DL_DATA_LEVEL',
                },
                'FILE_OUTPUT':
                {
                    'format': 'outformat',
                    'style': 'outstyle',
                    'createfiles': 'createfiles',
                    'destdir': 'destdir',
                    'xmlwrap': 'xmlwrap',
                    'flags': 'flags',
                    'fileindex': 'indexdb',
                    'unitconvert_write': 'gen_converted',
                    'long_timeseries': 'long_timeseries',
                }
            }
        if custom_parser == 'NASA_READ':
            # keyword parameters for EbasNasaAmes init:
            ret['nas_init'] = DataObject()
            ret['nas_init']['condense_messages'] = self.args.condense_messages
            # keyword parameters for EbasNasaAmes read:
            ret['nas_read'] = DataObject()
            ret['nas_read']['ignore_numformat'] = False
            if 'numformat' in self.args.ignore:
                ret['nas_read']['ignore_numformat'] = True
            ret['nas_read']['ignore_dx0'] = False
            if 'dx0' in self.args.ignore:
                ret['nas_read']['ignore_dx0'] = True
            ret['nas_read']['ignore_revdate'] = False
            if 'revdate' in self.args.ignore:
                ret['nas_read']['ignore_revdate'] = True
            ret['nas_read']['ignore_rescode'] = False
            if 'rescode' in self.args.ignore:
                ret['nas_read']['ignore_rescode'] = True
            ret['nas_read']['ignore_sampleduration'] = False
            if 'sample-duration' in self.args.ignore:
                ret['nas_read']['ignore_sampleduration'] = True
            ret['nas_read']['ignore_identicalvalues'] = False
            if 'identical-values' in self.args.ignore:
                ret['nas_read']['ignore_identicalvalues'] = True
            ret['nas_read']['ignore_parameter'] = False
            if 'parameter' in self.args.ignore:
                ret['nas_read']['ignore_parameter'] = True
            ret['nas_read']['ignore_flagconsistency'] = False
            if 'flag-consistency' in self.args.ignore:
                ret['nas_read']['ignore_flagconsistency'] = True
            ret['nas_read']['ignore_valuecheck'] = False
            if 'value-check' in self.args.ignore:
                ret['nas_read']['ignore_valuecheck'] = True
            if 'template-check' in self.args.ignore:
                ret['nas_read']['ignore_templatecheck'] = True
            ret['nas_read']['fix_flagconsistency'] = False
            if 'flag-consistency' in self.args.fix:
                ret['nas_read']['fix_flagconsistency'] = True
            ret['nas_read']['fix_overlap_set_start'] = False
            if 'overlap_set_start' in self.args.fix:
                ret['nas_read']['fix_overlap_set_start'] = True
            ret['nas_read']['fix_bc211'] = False
            if 'bc211' in self.args.fix:
                ret['nas_read']['fix_bc211'] = True
            ret['nas_read']['skip_variables'] = \
                [self.args.skip_variables] \
                    if isinstance(self.args.skip_variables, int) \
                    else self.args.skip_variables
            ret['nas_read']['skip_data'] = self.args.skip_data
            # reversed logic for skip_unitconvert
            ret['nas_read']['skip_unitconvert'] = not self.args.unitconvert_read
        elif custom_parser == 'OTHER_READ':
            # keyword parameters for EbasNasaAmes init:
            ret['file_init'] = DataObject()
            # keyword parameters for EbasNasaAmes read:
            ret['file_read'] = DataObject()
            ret['file_read']['ignore_identicalvalues'] = False
            if 'identical-values' in self.args.ignore:
                ret['file_read']['ignore_identicalvalues'] = True
            ret['file_read']['ignore_parameter'] = False
            if 'parameter' in self.args.ignore:
                ret['file_read']['ignore_parameter'] = True
            ret['file_read']['ignore_flagconsistency'] = False
            if 'flag-consistency' in self.args.ignore:
                ret['file_read']['ignore_flagconsistency'] = True
            ret['file_read']['ignore_valuecheck'] = False
            if 'value-check' in self.args.ignore:
                ret['file_read']['ignore_valuecheck'] = True
            # reversed logic for skip_unitconvert
            ret['file_read']['skip_unitconvert'] = not self.args.unitconvert_read
        else:
            for argname, critname in list(translate[custom_parser].items()):
                if argname in self.args.__dict__ and \
                        self.args.__dict__[argname] is not None:
                    ret[critname] = self.args.__dict__[argname]
        # additional elemet for DS_CRIT: add single characteristics
        if custom_parser == 'DS_CRIT' and self.args.characteristics:
            for key in self.args.characteristics.keys():
                ret[key] = self.args.characteristics[key]
        return ret

class ValidateArgument(ParseStrings):
    """
    Argument validation. Based on ParseStrings.
    """

    def __init__(self, ebasargs, validation):
        ParseStrings.__init__(self, allow_scalar=True, allow_list=True)
        self.ebasargs = ebasargs
        self.validation = validation

    def __call__(self, arguments):
        """
        Validation for station code or station code list.
        Parameters:
            string   cmd-line parameter string to be checked
        Reurns:
            string   in case of a single station code
            [string, ...]
                     in case of a comma separated list of station codes
        Raises:
            argparse.ArgumentTypeError
                     in case of argument error
        """
        return self.validation(self, arguments)

    def val_stationcode(self, station_codes):
        """
        Validation for station codes.
        Parameters:
            station_codes   cmd-line parameter string to be checked
        Reurns:
            string   in case of a single station code
            [string, ...]
                     in case of a comma separated list of station codes
        Raises:
            argparse.ArgumentTypeError
                     in case of argument error
        """
        def _check_stationcode(code):
            """
            Checks single station coode fromcommand line.
            Parameters:
                code     station code as entered at commandline
            Returns:
                code     modified station code (maybe appended '%' wildcard)
            Raises:
                argparse.ArgumentTypeError in case of syntax error or no
                matching station
            """
            if re.search('%', code):
                search = code
                search = re.sub('%', '.*', search)
            elif re.match('[A-Z][A-Z]$', code) or \
                 re.match('[A-Z][A-Z][0-9][0-9][0-9][0-9]$', code):
                search = code + '.*'
                code += '%'
            elif re.match('[A-Z][A-Z][0-9][0-9][0-9][0-9][A-Z]$', code):
                search = code
            else:
                raise argparse.ArgumentTypeError(
                    "station code syntax error '{}'".format(code))
            if len(list(st_.lookup_stationcode(search))) < 1:
                raise argparse.ArgumentTypeError(
                    "No match for station code '{}'".format(code))
            return code


        from ebas.domain.masterdata.st import EbasMasterST
        st_ = EbasMasterST(self.ebasargs.dbh)
        arg = self.parse(station_codes)
        if isinstance(arg, six.string_types):
            res = _check_stationcode(arg)
            return res
        elif isinstance(arg, list):
            for i in range(len(arg)):
                arg[i] = _check_stationcode(arg[i])
            return arg
        else:
            raise argparse.ArgumentTypeError('station code or station code '
                                             'list expected')

    def val_component(self, comp_name):
        """
        Validation for component or or component list.
        Parameters:
            string   cmd-line parameter string to be checked
        Reurns:
            string   in case of a single component
            [string, ...]
                     in case of a comma separated list of components
        Raises:
            argparse.ArgumentTypeError
                     in case of argument error
        """
        from ebas.domain.masterdata.cy import EbasMasterCY
        from ebas.domain import EbasDomainError
        cy_ = EbasMasterCY(self.ebasargs.dbh)
        arg = self.parse(comp_name)
        if isinstance(arg, six.string_types):
            try:
                return cy_.lookup_synonym_strict(arg, case_sensitive=False)
            except EbasDomainError as expt:
                raise argparse.ArgumentTypeError(str(expt))
        elif isinstance(arg, list):
            for i, string in enumerate(arg):
                try:
                    arg[i] = cy_.lookup_synonym_strict(string,
                                                       case_sensitive=False)
                except EbasDomainError as expt:
                    raise argparse.ArgumentTypeError(str(expt))
            return arg
        else:
            raise argparse.ArgumentTypeError('component or component list '
                                             'expected')

    def val_project(self, projects):
        """
        Validation for project or project list.
        Parameters:
            projects   cmd-line parameter string to be checked
        Reurns:
            string   in case of a single project
            [string, ...]
                     in case of a comma separated list of projects
        Raises:
            argparse.ArgumentTypeError
                     in case of argument error
        """
        from ebas.domain.masterdata.pr import EbasMasterPR
        from ebas.domain import EbasDomainError
        pr_ = EbasMasterPR(self.ebasargs.dbh)
        arg = self.parse(projects)
        if isinstance(arg, six.string_types):
            try:
                return pr_.lookup_project(arg)
            except EbasDomainError as expt:
                raise argparse.ArgumentTypeError(str(expt))
        elif isinstance(arg, list):
            for i, string in enumerate(arg):
                try:
                    arg[i] = pr_.lookup_project(string)
                except EbasDomainError as expt:
                    raise argparse.ArgumentTypeError(str(expt))
            return arg
        else:
            raise argparse.ArgumentTypeError('project or project list '
                                             'expected')

    def val_datalevel(self, datalevel):
        """
        Validation for datalevel.
        Parameters:
            datalevel   cmd-line parameter string to be checked
        Reurns:
            string   datalevel, checked and OK
        Raises:
            argparse.ArgumentTypeError
                     in case of argument error
        """
        from ebas.domain.masterdata.dl import EbasMasterDL
        dl_ = EbasMasterDL(self.ebasargs.dbh)
        arg = self.parse(datalevel)
        if isinstance(arg, six.string_types):
            try:
                return dl_[arg].DL_DATA_LEVEL
            except KeyError:
                raise argparse.ArgumentTypeError(
                    "Datalevel '{}' not defined".format(arg))
        elif isinstance(arg, list):
            for i, string in enumerate(arg):
                try:
                    arg[i] = dl_[string].DL_DATA_LEVEL
                except KeyError:
                    raise argparse.ArgumentTypeError(
                        "Datalevel '{}' not defined".format(string))
            return arg
        else:
            raise argparse.ArgumentTypeError(
                'datalevel or datalevel list expected')

    def val_instrument(self, instruments):
        """
        Validation for instrument type or instrument type list.
        Parameters:
            instruments   cmd-line parameter string to be checked
        Reurns:
            string   in case of a single instrument type
            [string, ...]
                     in case of a comma separated list of instrument types
        Raises:
            argparse.ArgumentTypeError
                     in case of argument error
        """
        arg = self.parse(instruments)
        return arg
        # TODO ...

    def val_matrix(self, matrices):
        """
        Validation for matrix or matrix list.
        Parameters:
            matrices   cmd-line parameter string to be checked
        Reurns:
            string   in case of a single matrix
            [string, ...]
                     in case of a comma separated list of matrices
        Raises:
            argparse.ArgumentTypeError
                     in case of argument error
        """
        arg = self.parse(matrices)
        return arg
        # TODO ...

    def val_group(self, groups):
        """
        Validation for parameter group.
        Parameters:
            groups   cmd-line parameter string to be checked
        Reurns:
            string   in case of a group
            [string, ...]
                     in case of a comma separated list of groups
        Raises:
            argparse.ArgumentTypeError
                     in case of argument error
        """
        arg = self.parse(groups)
        return arg
        # TODO ...

    def val_fi_ref(self, fi_refs):
        """
        Validation for instrument references.
        Parameters:
            fi_refs   cmd-line parameter string to be checked
        Reurns:
            string   in case of one fi_ref
            [string, ...]
                     in case of a comma separated list of fi_refs
        Raises:
            argparse.ArgumentTypeError
                     in case of argument error
        """
        arg = self.parse(fi_refs)
        return arg
        # TODO ...

    def val_me_ref(self, me_refs):
        """
        Validation for method references.
        Parameters:
            me_refs   cmd-line parameter string to be checked
        Reurns:
            string   in case of one me_ref
            [string, ...]
                     in case of a comma separated list of me_refs
        Raises:
            argparse.ArgumentTypeError
                     in case of argument error
        """
        arg = self.parse(me_refs)
        return arg
        # TODO ...

    def val_resolution(self, rescodes):
        """
        Validation for resolution codes.
        Parameters:
            rescodes   cmd-line parameter string to be checked
        Reurns:
            string   in case of one resolution code
            [string, ...]
                     in case of a comma separated list of resolution codes
        Raises:
            argparse.ArgumentTypeError
                     in case of argument error
        """
        arg = self.parse(rescodes)
        return arg
        # TODO ...

    def val_statistics(self, statistics):
        """
        Validation for statistic codes.
        Parameters:
            statistics   cmd-line parameter string to be checked
        Reurns:
            string   in case of one statistic code
            [string, ...]
                     in case of a comma separated list of statistic codes
        Raises:
            argparse.ArgumentTypeError
                     in case of argument error
        """
        arg = self.parse(statistics)
        return arg
        # TODO ...

    def val_characteristics(self, characteristics):
        """
        Validation for statistic codes.
        Parameters:
            characteristics   cmd-line parameter string to be checked
        Reurns:
            string   in case of one statistic code
            [string, ...]
                     in case of a comma separated list of statistic codes
        Raises:
            argparse.ArgumentTypeError
                     in case of argument error
        """
        arg = self.parse(characteristics)
        ret = {}
        for char in arg.split(','):
            tag, val = char.split('=')
            ret[tag] = val
        return ret

def option_bool(classname, truevalue, falsevalue):
    """
    Class factory, only used to translate bool value to text and vice versa.
    """
    class OptionBool(int):
        """
        The class itself, needs to be int based (bool cannot be used as base
        class).
        """
        TRUEVALUE = truevalue
        FALSEVALUE = falsevalue

        def __new__(cls, val):
            if val == cls.TRUEVALUE:
                return super(OptionBool, cls).__new__(cls, True)
            if val == cls.FALSEVALUE:
                return super(OptionBool, cls).__new__(cls, False)
            return super(OptionBool, cls).__new__(cls, val)
        def __repr__(self):
            return "{}({})".format(self.__class__.__name__, str(self))
        def __str__(self):
            if self:
                return self.__class__.TRUEVALUE
            return self.__class__.FALSEVALUE
    OptionBool.__name__ = classname
    OptionBool.__qualname__ = classname
    return OptionBool

def option_int(classname, text_to_val):
    """
    Class factory, only used to translate bool value to text and vice versa.
    """
    class OptionInt(int):
        """
        The class itself.
        """
        TEXT_TO_VAL = text_to_val
        VAL_TO_TEXT = {text_to_val[x]: x for x in text_to_val.keys()}

        def __new__(cls, text):
            return super(OptionInt, cls).__new__(
                cls, cls.TEXT_TO_VAL[text])
        def __repr__(self):
            return "{}({})".format(self.__class__.__name__, str(self))
        def __str__(self):
            return self.__class__.VAL_TO_TEXT[self]
    OptionInt.__name__ = classname
    OptionInt.__qualname__ = classname
    return OptionInt

class ParseDestDirAction(argparse.Action):  # pylint: disable-msg=R0903
    # R0903: ParseDestDirAction: Too few public methods
    #  this is due to argparse.Action
    """
    Parser action for --destdir argument: is only allowed when --createfiles is
    set.
    """
    def __call__(self, parser, namespace, values, _option_string=None):
        createfiles = getattr(namespace, 'createfiles', False)
        if not createfiles:
            parser.error("--destdir is only allowed after --createfiles!")
        else:
            setattr(namespace, self.dest, values)

def parse_fileindex(string):
    """
    Parser for fileindex. Returns a IndexDb object.

    Parameters:
       string   cmd-line parameter string to be parsed.
    Returns:
       IndexDb object
    Raises:
        argparse.ArgumentTypeError
    """
    if not string:
        return None
    append = False
    if string.startswith('+'):
        append = True
        string = string[1:]
    try:
        indexdb = IndexDb(string)
    except OperationalError as excpt:
        if excpt.message == 'unable to open database file':
            raise argparse.ArgumentTypeError("can not open index db {}".format(
                string))
        else:
            raise
    try:
        indexdb.create_tables()
    except OperationalError as excpt:
        if re.match('table .* already exists', excpt.message):
            if not append:
                raise argparse.ArgumentTypeError(
                    "index db {} exists already".format(string))
            # if append, ignore
        else:
            raise
    return indexdb


# output routines shared by different programs:

def dslist_info(dslist):
    """
    Interactive information on a result ds list.
    Parameters:
        dslist
    Returns:
        None
    """
    t_overv = "Overview of different metadata"
    t_list = "List of all dataset metadata"
    t_verb = "Verbose information on all datasets"
    t_sverb = "Super-verbose information on all datasets"
    t_exit = "Exit dataset information"
    choice = ''
    while choice != 'Exit':
        choice = input("\nPlease make your choice:\n " + \
            t_overv + "\n " + t_list + "\n " + t_verb + "\n " + \
            t_sverb + "\n " + t_exit + "\n ")
        print("")
        if t_overv.lower().startswith(choice.lower()):
            for str_ in dslist.metadata_to_string(verbosity=-1):
                print(str_)
        if t_list.lower().startswith(choice.lower()):
            for str_ in dslist.metadata_to_string(verbosity=0):
                print(str_)
        elif t_verb.lower().startswith(choice.lower()):
            for str_ in dslist.metadata_to_string(verbosity=1):
                print(str_)
        elif t_sverb.lower().startswith(choice.lower()):
            for str_ in dslist.metadata_to_string(verbosity=2):
                print(str_)
        if t_exit.lower().startswith(choice.lower()):
            choice = "Exit"

def _confirmation_issues(dslist, verb):
    """
    Check if confirmation is needed and log warnings about thresholds exeeded.

    Confirmation is needed if there is a high risk that the user made
    a mistake when generating the commandline arguments (number of datasets,
    multiple instruments...).
    Parameters:
        dslist       list of EbasDomDS objects to be affected
        interactive  user interactive mode (bool)
        verb         verb to be used for the action intended (delete,
                     change, ...)
                     The verb needs to be defined in two different grammatical
                     forms, passed as a tuple:
                         (second person singular simple present tense active,
                          third person plural future tense passive)
                     Examples: (delete, deleted), (change, changed)
    Returns:
        issues  number of issues discovered (>0 --> confirmation is needed)
    """
    issues = []

    stations = set([x.ST_STATION_CODE for x in dslist])
    if len(stations) > 1:
        issues.append("You're about to {} data from {} stations: {}"
                      .format(verb[0], len(stations),
                              ", ".join(sorted(list(stations)))))
    instruments = set([x.FI_REF for x in dslist])
    if len(instruments) > 1:
        issues.append("You're about to {} data from {} instruments: {}"
                      .format(verb[0], len(instruments),
                              ", ".join(sorted(list(instruments)))))
    nrt = [ds.NRT for ds in dslist if ds.NRT]
    if nrt:
        issues.append("You're selection contains {} NRT datasets. "
                      .format(len(nrt)) +
                      "Data will be {} permanently.".format(verb[1]))
    return issues

def confirm_changes(dslist, interactive, verb=('delete', 'deleted')):
    """
    Generates warnings if there is a high risk that the user made a
    mistake when generating the commandline arguments (number of datasets,
    multiple instruments...).

    Parameters:
        dslist       list of EbasDomDS objects to be affected
        interactive  user interactive mode (bool)
        verb         verb to be used for the action intended (delete,
                     change, ...)
                     The verb needs to be defined in two different grammatical
                     forms, passed as a tuple:
                         (second person singular simple present tense active,
                          third person plural future tense passive)
                     Examples: (delete, deleted), (change, changed)
    Returns:
        True: go ahead
        False: stop
    """
    logger = logging.getLogger('EbasCmdLine')
    messages = _confirmation_issues(dslist, verb)
    for msg in messages:
        if interactive:
            logger.warning(msg)
        else:
            logger.error(msg)
    issues = len(messages)
    if issues and interactive:
        choice = ''
        while choice not in ('yes', 'no'):
            yes = "Yes, I am sure. Please go ahead!"
            no_ = "No, actually I am not so sure. Please get me out here!"
            info = "I am hesitant. Please give me more information!"
            choice = input(
                "\nPlease read the {}{} of warning above. "
                .format(issues if issues > 1 else '',
                        " lines" if issues > 1 else "line") +
                "Are you sure you want to {}?\n ".format(verb[0]) +
                yes + "\n " + no_ + "\n " + info + "\n ")
            if yes.lower().startswith(choice.lower()):
                choice = "yes"
            if no_.lower().startswith(choice.lower()):
                choice = "no"
            if info.lower().startswith(choice.lower()):
                dslist_info(dslist)
        return choice == "yes"
    elif issues:
        logger.info('User interaction impossible, comntact the EBAS '
                    'admin team')
        return False
    # no issues:
    return True




"""
ebas/io/base.py
$Id: check_data.py 2773 2022-01-08 02:29:35Z pe $

Base classes for (file) I/O.

"""

import datetime
from nilutility.datetime_helper import datetime_round
from nilutility.list_helper import all_none, all_equal
from ebas.domain.masterdata.bc import EbasMasterBC, \
    EbasMasterBCNoProj, EbasMasterBCDuplicateErrorList
from ebas.domain.basic_domain_logic.flags import check_value_flag_consistency, \
    ISSUE_SEVERITY_ERROR
from ebas.domain.basic_domain_logic.value_check import check_values
from ebas.domain.basic_domain_logic.datalevel import base_datalevel
from ebas.domain.masterdata.pg_pl import EbasMasterPL
from .base import EbasFileBase, isvalid


class EbasFileCheckData(EbasFileBase):  # pylint: disable=W0223
    # W0223: ... is abstract in base class but is not overridden
    # this class is abstract as well (does not have any NotImplementedError)
    """
    Partial class for EbasFile (Base class for file objects).
    This part handles checking of the data part.
    """

    def check_data(
        self, ignore_identicalvalues=False,
        ignore_flagconsistency=False, fix_flagconsistency=False,
        ignore_nodata=False, ignore_valuecheck=False,
        fix_overlap_set_start=False, fix_bc211=False):
        """
        Check data consistency and any cause for suspicion in the data.
        Parameter:
            ignore_identicalvalues
                    ignore errors related to all values of one
                    variable being equal (e.g. all data missing for
                    one variable, all values 0.0 or all values any
                    other constant value).
            ignore_flagconsistency
                    ignore errors regarding flag consistency and flag/value
                    consitency
            ignore_nodata
                    the object is expected to not contain any data (ie.g. read
                    with skip_data)
            ignore_valuecheck ignore value checks (boundaries and spikes)
            fix_overlap_set_start
                    instead of raising an error oin case of overlapping samples,
                    the starttime i set to the previous end time
            fix_bc211
                    flag boundary check violations with flag 211
        """
        self._check_sample_times(ignore_nodata, fix_overlap_set_start)
        self._check_value_flag_consistency(
            ignore_flagconsistency, fix_flagconsistency)
        # boundaries and spikes:
        self._check_values(ignore_valuecheck, fix_bc211=fix_bc211)
        self._check_identical_values(ignore_identicalvalues)

    def _check_sample_times(self, ignore_nodata, fix_overlap_set_start):
        """
        Parameters:
            ignore_nodata
                    the object is expected to not contain any data (ie.g. read
                    with skip_data)
            fix_overlap_set_start
                    instead of raising an error oin case of overlapping samples,
                    the starttime i set to the previous end time
        """
        _unused, file_lnum, _ = self.msgref_vnum_lnum(0)
        if not self.msg_condenser.is_reset():
            raise RuntimeError('Uninitialized MessageCondenser')
        for i in range(len(self.sample_times)):
            if i > 0 and (self.sample_times[i-1][1] - \
                          self.sample_times[i][0]).total_seconds() > 0.0:
                if fix_overlap_set_start:
                    self.msg_condenser.add(
                        self.warning, self.msg_condenser.MSG_ID_TIME_OVERLAP,
                        None,
                        "start time {} is not >= previous end time {}, fixing "
                        "start time".format(
                            self.sample_times[i][0], self.sample_times[i-1][1]),
                        file_lnum+i)
                    self.sample_times[i][0] = self.sample_times[i-1][1]
                    for var in self.variables:
                        if 247 not in var.flags[i]:
                            # flag 247: Overlapping sample interval was
                            # corrected by the database co-ordinator. Possible
                            # wrong sample time (used for historic data only)
                            var.flags[i].append(247)
                else:
                    self.msg_condenser.add(
                        self.error, self.msg_condenser.MSG_ID_TIME_OVERLAP,
                        None,
                        "start time {} is not >= previous end time {}".format(
                            self.sample_times[i][0], self.sample_times[i-1][1]),
                        file_lnum+i)
            if not self.sample_times[i][0] < self.sample_times[i][1]:
                self.msg_condenser.add(
                    self.error, self.msg_condenser.MSG_ID_TIME_INVERT,
                    None, "starttime is not < endtime", file_lnum+i)
        # write all condensed messages and reset condenser object:
        self.msg_condenser.deliver()

        # check first starttime is equal metadata elemen Startdate
        self._check_startdate(ignore_nodata, file_lnum)

        # check last endtime is not in the future.
        # This is only true for observed data (not for model data).
        # Needs reconsideration when we accept regimes other than IMG
        now = datetime.datetime.utcnow()
        if self.sample_times and \
           self.sample_times[-1][1] is not None and \
           self.sample_times[-1][1] >= now:
            self.error(
                "No future data allowed. Last sample endtime is '{}'.".\
                format(self.sample_times[-1][1].strftime("%Y%m%d%H%M%S")))

        if not self.sample_times and not ignore_nodata:
            # reason to report an error when a file contains no (valid) data:
            # revdate and period code can not be checked, which means the header
            # cannot be validated without data, thus a file without data cannot
            # be valid
            self.error('File contains no data')

    def _check_startdate(self, ignore_nodata, lnum):
        """
        Check the metadata elemet startdate against first sample start.
        Parameters:
            firststart   starttime of first sample
            lnum         line number (of first data line) (for error reporting)
        Returns:
            None
        """
        if not self.metadata.startdate:
            self.error("No start date set in metadata")
        if not ignore_nodata and (\
                not self.sample_times or not self.sample_times[0] or \
                not self.sample_times[0][0]):
            # report error only if not ignore_nodata
            self.error(
                "No samples in file, cannot check metadata start date",
                lnum=lnum)

        if not self.metadata.startdate or not  self.sample_times or \
                not self.sample_times[0] or not self.sample_times[0][0]:
            return

        firststart = self.sample_times[0][0]
        if self.metadata.datadef == 'EBAS_1' and \
           self.metadata.startdate.time() == datetime.time(0, 0):
            # EBAS_1 and startdate is only specified with date precision, only
            # check for same day.
            if self.metadata.startdate.date() != firststart.date():
                self.error(
                    "start day of first sample ({}) is not equal to "
                    "'Startdate' specified in metadata ({})".format(
                        firststart.strftime('%Y-%m-%d'),
                        self.metadata.startdate.strftime('%Y-%m-%d')),
                    lnum=lnum)

        elif abs((firststart-self.metadata.startdate).total_seconds()) >= 1800:
            # difference >= 30 minutes
            self.error(
                "start time of first sample ({}) is not equal to 'Startdate' "
                "specified in metadata ({})".format(
                    datetime_round(firststart, 0).strftime('%Y-%m-%dT%H:%M:%S'),
                    self.metadata.startdate.strftime('%Y-%m-%dT%H:%M:%S')),
                lnum=lnum)
        elif abs((firststart-self.metadata.startdate).total_seconds()) >= 0.5:
            # difference >= 0.5 sec
            self.warning(
                "start time of first sample ({}) is not equal to 'Startdate' "
                "specified in metadata ({})".format(
                    datetime_round(firststart, 0).strftime('%Y-%m-%dT%H:%M:%S'),
                    self.metadata.startdate.strftime('%Y-%m-%dT%H:%M:%S')),
                lnum=lnum)

    def _check_value_flag_consistency(self, ignore_flagconsistency,
                                      fix_flagconsistency):
        """
        Check valididy of flags, consistency of flags and flag/value
        combinations.
        Parameters:
            ignore_flagconsistency
                    ignore errors regarding flag consistency and flag/value
                    consitency
        Returns:
            None
        """
        if not self.msg_condenser.is_reset():
            raise RuntimeError('Uninitialized MessageCondenser')
        for vnum in range(len(self.variables)):  # pylint: disable=C0200
            # C0200: Consider using enumerate instead of iterating with
            # range/len
            # --> only using index
            file_vnum, file_lnum, _ = self.msgref_vnum_lnum(vnum)

            # Make sure the message condenser is fresh
            if not self.msg_condenser.is_reset():
                raise RuntimeError('Uninitialized MessageCondenser')

            # check all flags/values consistency
            special = 0
            if self.is_precip_amount(vnum):
                special = 1
            if self.is_auxiliary(vnum):
                special = 2
            for issue in check_value_flag_consistency(
                    self.variables[vnum].values_,
                    self.variables[vnum].flags,
                    special, ignore_flagconsistency, fix_flagconsistency):
                self.msg_condenser.add(
                    self.error if issue['severity'] == ISSUE_SEVERITY_ERROR
                    else self.warning,
                    issue['msg_id'], file_vnum,
                    "variable[{}]: ".format(file_vnum) + issue['message'],
                    file_lnum + issue['row'])
            # write all condensed messages and reset condenser object:
            self.msg_condenser.deliver()

    def _check_identical_values(self, ignore_identicalvalues):
        """
        Check for all identical values (including all zero and all None).
        Parameter:
            ignore_identicalvalues
                    ignore errors related to all values of one
                    variable being equal (e.g. all data missing for
                    one variable, all values 0.0 or all values any
                    other constant value).
        """
        if len(self.sample_times) > 10:
            # Do not check if it's too little values.
            # There are also some exceptions in _identical_values_exception_mask
            # depending on number of samples or sample period.
            for vnum in range(len(self.variables)):
                exceptions = self._identical_values_exception_mask(vnum)
                file_vnum, _, _ = self.msgref_vnum_lnum(vnum)

                # some sanity checks:
                messagefunc = \
                    self.warning if ignore_identicalvalues else self.error
                if all_none(self.variables[vnum].values_):
                    if not exceptions & \
                            self.__class__.CHECK_EXCEPTION_ALLNONE_SILENT:
                        # only raise any message if not _SILENT, else skip
                        # message alltogether
                        if exceptions & self.__class__.CHECK_EXCEPTION_ALLNONE:
                            messagefunc = self.warning
                        messagefunc(
                            "variable[{}]: All values are MISSING. Check data "
                            "sanity.".format(file_vnum))
                elif all_equal(self.variables[vnum].values_, ignore_none=True):
                    # get first none None value:
                    value = next((x for x in self.variables[vnum].values_
                                  if x is not None))
                    message = ("variable[{}]: All valid values are {}. Check "
                               "data sanity.").format(file_vnum, value)
                    dl_ = self.get_meta_for_var(vnum, "detection_limit")
                    if value == 0.0:
                        if not exceptions & \
                                self.__class__.CHECK_EXCEPTION_ALLZERO_SILENT:
                            # only raise any message if not _SILENT, else skip
                            # message alltogether
                            if exceptions & \
                                    self.__class__.CHECK_EXCEPTION_ALLZERO:
                                messagefunc = self.warning
                            messagefunc(message)
                    else:
                        if not exceptions & \
                                self.__class__.CHECK_EXCEPTION_ALLEQUAL_SILENT:
                            # only raise any message if not _SILENT, else skip
                            # message alltogether
                            if exceptions & \
                                    self.__class__.CHECK_EXCEPTION_ALLEQUAL:
                                messagefunc = self.warning
                            # exception for BDL values: if only BDL values and
                            # all are set to BDL, this error needs to be
                            # downgraded to a warning:
                            if all([781 in self.variables[vnum].flags[i]
                                    for i in range( \
                                        len(self.variables[vnum].values_))
                                    if self.variables[vnum].values_[i] \
                                        is not None]):
                                message = (
                                    "variable[{}]: All valid values are {} (below "
                                    "detection limit, flagged 781)").format(
                                        file_vnum, value)
                                messagefunc = self.warning
                            messagefunc(message)

    CHECK_EXCEPTION_ALLEQUAL = 1  # allow equal, but not zero or None
    CHECK_EXCEPTION_ALLEQUAL_SILENT = 2  # addtn'lly, be silent in this case
    CHECK_EXCEPTION_ALLZERO = 4  # special case of identilcalvalues, needed
    CHECK_EXCEPTION_ALLZERO_SILENT = 8  # # addtn'lly, be silent in this case
    CHECK_EXCEPTION_ALLNONE = 16
    CHECK_EXCEPTION_ALLNONE_SILENT = 32 # addtn'lly, be silent in this case
    # often used sums:
    CHECK_EXCEPTION_ALLIDENTICAL = CHECK_EXCEPTION_ALLEQUAL | \
        CHECK_EXCEPTION_ALLZERO | CHECK_EXCEPTION_ALLNONE
        # allow equal, zero or None
    CHECK_EXCEPTION_ALLIDENTICAL_SILENT =  CHECK_EXCEPTION_ALLEQUAL_SILENT | \
        CHECK_EXCEPTION_ALLZERO_SILENT | CHECK_EXCEPTION_ALLNONE_SILENT
        # addtn'lly, be silent in this case

    def _identical_values_exception_mask(self, vnum):
        """
        Get possible exceptions for checking the variable.
        Parameters:
            vnum              variable index
            exception_mask    bitmask of exceptions to be checked
        """
        exceptions = 0
        instr_type = self.get_meta_for_var(vnum, "instr_type")
        regime = self.get_meta_for_var(vnum, "regime")
        matrix = self.get_meta_for_var(vnum, "matrix")
        comp_name = self.get_meta_for_var(vnum, "comp_name")
        data_level = base_datalevel(self.get_meta_for_var(vnum, 'datalevel'))
        charact = self.get_characteristics_for_var(vnum)
        statistics = self.get_meta_for_var(vnum, "statistics")

        # General exceptions
        if comp_name == 'status':
            exceptions |= self.__class__.CHECK_EXCEPTION_ALLIDENTICAL
        if 'Actual/target' in charact and charact['Actual/target'] == 'target':
            exceptions |= self.__class__.CHECK_EXCEPTION_ALLIDENTICAL
        if ('Nominal/measured' in charact and
            charact['Nominal/measured'] in ('nominal',
                                            'Calibration gas concentration')):
            exceptions |= self.__class__.CHECK_EXCEPTION_ALLEQUAL_SILENT
        if 'Assumption' in charact and charact['Assumption']:
            exceptions |= self.__class__.CHECK_EXCEPTION_ALLIDENTICAL
        if 'Actual/target' in charact and charact['Actual/target'] == 'target':
            exceptions |= self.__class__.CHECK_EXCEPTION_ALLIDENTICAL

        if comp_name == 'particle_number_size_distribution':
            # ops, size > 400 nm: ignore 0 values (the bigger sizes the more
            # oftern there are in fact no particles detected for a whole year.
            # depends on the station setting (Troll 400 nm, Greece 16000)
            # aps depends on station (e.g. KR0100R 19000 nm)
            # We make an exception for the upper 30% of sizes
            all_sizes = set()
            for i, var in enumerate(self.variables):
                if (var.metadata.comp_name ==
                    'particle_number_size_distribution'):
                    charact1 = self.get_characteristics_for_var(i)
                    if 'D' in charact1:
                        all_sizes.add(charact1['D'])
            if len(all_sizes) > 2:
                threshold = sorted(all_sizes)[int(2*len(all_sizes)/3)]
                if 'D' in charact and charact['D'] >= threshold:
                    exceptions |= self.__class__.CHECK_EXCEPTION_ALLZERO
            # MPSS, smallest size bin, lower percentile, allow all 0
            if (statistics == 'percentile:15.87' and
                'D' in charact and charact['D'] == sorted(all_sizes)[0]):
                exceptions |= self.__class__.CHECK_EXCEPTION_ALLZERO
        if instr_type == 'TEOM' and \
           comp_name == 'temperature':
            exceptions |= self.__class__.CHECK_EXCEPTION_ALLIDENTICAL
        if data_level == '0' and  instr_type in ('dmps', 'smps'):
            if comp_name in ('status', 'temperature', 'flow_rate',
                             'particle_diameter'):
                exceptions |= self.__class__.CHECK_EXCEPTION_ALLIDENTICAL
            if comp_name == 'particle_number_concentration' and \
               'Size bin' in charact and charact['Size bin'] > 25:
                # ther might be no "big" particles all year...
                # however we define big here in level 0 (let's say 25th bin)
                exceptions |= self.__class__.CHECK_EXCEPTION_ALLZERO
        if instr_type == 'aerosol_mass_spectrometer':
            if (data_level == '0' and
                comp_name in ('pressure', 'temperature', 'relative_humidity',
                              'flow_rate','status',
                              'frequency', 'electric_power',
                              'electric_tension', 'electric_current',
                              'airbeam_signal', 
                              'nitrogen_ion_flow_signal',
                              'ionization_efficiency',
                              'relative_ionization_efficiency',
                              'collection_efficiency')):
                exceptions |= self.__class__.CHECK_EXCEPTION_ALLIDENTICAL
            if comp_name == 'ammonium':
                # 2023-11-16: There is an issue with ammonium for ToF ACSM
                # baseline below 0, so data submitters flag out all ammonium
                # in lev0, which means all missing in lev2.
                # We accept this, or we would not get any data...
                # TODO: this is only temporary, as long as the underlying
                # problem is not fixed
                exceptions |= self.__class__.CHECK_EXCEPTION_ALLNONE
        if instr_type == 'filter_absorption_photometer' and \
           comp_name in ('flow_rate', 'bypass_fraction'):
            # the bypass fraction is usually set and left constant for all
            # measurements (it's instrument configuration, really)
            # Therefore, silence this issue completely (not even a warning)
            exceptions |= self.__class__.CHECK_EXCEPTION_ALLIDENTICAL_SILENT
        if instr_type == 'filter_absorption_photometer' and \
           data_level == '0':
            if comp_name in ('status', 'temperature', 'pressure'):
                # AE33 temp and pressure is a reference value (for volume
                # standard) and NOT measured, thus it can be constant.
                # TODO: change AE33 template, use Assumption=reference pressure
                # Some of the status variables may be constant over a year...
                exceptions |= self.__class__.CHECK_EXCEPTION_ALLIDENTICAL
            if comp_name == 'filter_loading_compensation_parameter' and \
                    (self.sample_times[-1][1]-self.sample_times[0][0]).\
                    total_seconds() <= 3600*24*10:
                # AE33 filter_loading_compensation_parameter may be constant
                # over some time time?
                exceptions |= self.__class__.CHECK_EXCEPTION_ALLIDENTICAL
            if comp_name == 'filter_number' and \
                    (self.sample_times[-1][1]-self.sample_times[0][0]).\
                    total_seconds() <= 3600*24*30*3:
                # AE33 filter_number may be constant over 3 month
                exceptions |= self.__class__.CHECK_EXCEPTION_ALLIDENTICAL
        if instr_type == 'filter_absorption_photometer' and \
           data_level == '0' and  \
           comp_name in ('relative_humidity', ):
            # AE42 no temp, pressure and humidity, but using template for AE31
            exceptions |= self.__class__.CHECK_EXCEPTION_ALLNONE

        if [1 for x in EbasMasterPL().list_groups(regime, matrix, comp_name)
                if x.startswith("VOC-") or x.startswith("voc_")]:
            # VOCs: precision and expanded uncertainty 2 sigma may turn out
            # constant if all measurements very low and around DL
            if statistics in ('precision', 'expanded uncertainty 2sigma'):
                exceptions |= self.__class__.CHECK_EXCEPTION_ALLEQUAL

        if [1 for x in EbasMasterPL().list_groups(regime, matrix, comp_name)
                if x == 'pollen']:
            # Pollen data: certain species can be 0 for long periods, maybe
            # always. This can be valueable information.
            exceptions |= self.__class__.CHECK_EXCEPTION_ALLZERO

        if instr_type == 'online_gc' and data_level == '0':
            # all components may be missing (not measured)
            exceptions |= self.__class__.CHECK_EXCEPTION_ALLNONE_SILENT
            if comp_name.endswith('_retention_time'):
                exceptions |= self.__class__.CHECK_EXCEPTION_ALLEQUAL
        if instr_type == 'online_gc' and data_level == '1':
            # all components may be missing (not measured)
            exceptions |= self.__class__.CHECK_EXCEPTION_ALLNONE_SILENT
        return exceptions

    def _check_values(self, ignore_valuecheck, fix_bc211=False):
        """
        Checks boundaries and spikes in the data of an EBAS NasaAmes Object.
        Parameters:
            ignore_valuecheck ignore value checks (boundaries and spikes)
            fix_bc211         flag boundary check violations with flag 211
        Returns:
            None
        """
        ebc = EbasMasterBC()
        trans_meta = [
            ('regime', 'RE_REGIME_CODE', 'Regime code'),
            ('matrix', 'MA_MATRIX_NAME', 'Matrix name'),
            ('comp_name', 'CO_COMP_NAME', 'Component name'),
            ('projects', 'PR_ACRONYM', 'Frameworks'),
            ('station_code', 'ST_STATION_CODE', 'Station code'),
            ('instr_type', 'FT_TYPE', 'Intrument type'),
            # ('datalevel', 'DL_DATA_LEVEL', 'Data level'),
            # Datalevel is not mandatory, we don't want to raise an error
            # if it is missing in the file!
            # Datalevel is handled separately below
            ('statistics', 'SC_STATISTIC_CODE', 'Statistics')]
        for vnum in range(len(self.variables)):  # pylint: disable=C0200
            # C0200: Consider using enumerate instead of iterating with
            # range/len
            # --> only using index
            undef = []
            search = {}
            file_vnum, file_lnum, _ = self.msgref_vnum_lnum(vnum)

            # translate metadata
            for elem in trans_meta:
                val = self.get_meta_for_var(vnum, elem[0])
                if not isvalid(val):
                    undef.append(elem[2])
                else:
                    search[elem[1]] = val

            # special case for datalevel: no error message when not set
            val = self.get_meta_for_var(vnum, 'datalevel')
            if isvalid(val):
                search['DL_DATA_LEVEL'] = val
            else:
                search['DL_DATA_LEVEL'] = None

            # special case for fi_ref:
            lab_code = self.get_meta_for_var(vnum, 'lab_code')
            instr_name = self.get_meta_for_var(vnum, 'instr_name')
            if not isvalid(lab_code):
                undef.append('Laboratory code')
            if not isvalid(instr_name):
                undef.append('Instrument name')
            if isvalid(lab_code) and isvalid(instr_name):
                search['FI_REF'] = lab_code + '_' + instr_name

            # any undefied metadata?
            if undef:
                self.error(
                    "variable[{}]: No applicable boundary check found because "
                    "some metadata are not specified correctly ({})".format(
                        file_vnum, ', '.join(undef)),
                    None)
                continue

            # characteristics:
            search['characteristics'] = self.get_characteristics_for_var(vnum)

            try:
                bc_list = ebc.select(search)
            except EbasMasterBCNoProj as excpt:
                self.error(
                    "variable[{}]: {}; boundary and spike checks skipped"
                    .format(file_vnum, excpt), None)
                bc_list = []
            except EbasMasterBCDuplicateErrorList as excpt:
                for err in excpt.errlist:
                    self.error(
                        "variable[{}]: {}; some boundary and spike checks "
                        "skipped".format(file_vnum, err), None)
                bc_list = excpt.retlist
            if not bc_list:
                self.warning(
                    "variable[{}]: No applicable boundary check found".format(
                        file_vnum), None)
            else:
                # check all rules:
                for rule in bc_list:
                    # Make sure the message condenser is fresh
                    if not self.msg_condenser.is_reset():
                        raise RuntimeError('Uninitialized MessageCondenser')
                    # check all values for variable vnum, filter by flags!
                    issues = check_values(
                        rule.PR_ACRONYM,
                        self.variables[vnum].values_,
                        self.variables[vnum].flags,
                        rule.BC_MINVAL, rule.BC_MAXVAL,
                        rule.BC_SPIKE_RADIUS,
                        rule.BC_SPIKE_MINOFFSET, rule.BC_SPIKE_MAXOFFSET,
                        rule.BC_SPIKE_MINSTDDEVFACT,
                        rule.BC_SPIKE_MAXSTDDEVFACT,
                        fix_bc211=fix_bc211)
                    for issue in issues:
                        self.msg_condenser.add(
                            self.error if not ignore_valuecheck and \
                                issue['severity'] == ISSUE_SEVERITY_ERROR \
                                else self.warning,
                            issue['msg_id'], file_vnum,
                            "variable[{}]: {}".format(file_vnum,
                                                      issue['message']),
                            file_lnum + issue['row'])
                    # write all condensed messages and reset condenser object:
                    self.msg_condenser.deliver()



"""
$Id: base.py 2427 2020-03-19 23:09:09Z pe $
"""

from nilutility.datatypes import ControlledDictList


ISSUE_SEVERITY_ERROR = 1
ISSUE_SEVERITY_WARNING = 2

###
### Message IDs (centralized here, to make sure IDs are unique
###

# flag issues, used in ebas.domain.basic_domain_logic.flags
MGGID_OFFSET_FLAGCHECK = 0
MSGID_FLAGCHECK_ILLEGAL = MGGID_OFFSET_FLAGCHECK + 1
MSGID_FLAGCHECK_MISSING_100 = MGGID_OFFSET_FLAGCHECK + 2
# flag issues based on checking flags/value cjnsistency or considering
# properties of the time series (precipitation amount, axiliary data)
# (checked in check_value_flag_consistency):
MSGID_FLAGCHECK_VALID_VALUE_MISSIG_FLAG = MGGID_OFFSET_FLAGCHECK + 3
MSGID_FLAGCHECK_MISSING_VALUE_FLAG_100 = MGGID_OFFSET_FLAGCHECK + 4
MSGID_FLAGCHECK_MISSING_VALUE_NO_FLAG = MGGID_OFFSET_FLAGCHECK + 5
MSGID_FLAGCHECK_PRECIP_FLAG_782_783_784_DROPPED = MGGID_OFFSET_FLAGCHECK + 6
MSGID_FLAGCHECK_PRECIP_FLAG_782_783_784_ERROR = MGGID_OFFSET_FLAGCHECK + 7
MSGID_FLAGCHECK_PRECIP_FLAG_890_DROPPED = MGGID_OFFSET_FLAGCHECK + 8
MSGID_FLAGCHECK_PRECIP_FLAG_890_SET_ZERO = MGGID_OFFSET_FLAGCHECK + 9
MSGID_FLAGCHECK_PRECIP_FLAG_890_ERROR = MGGID_OFFSET_FLAGCHECK + 10
MSGID_FLAGCHECK_AUX_VALID_VALUE_MISSING_FLAG = MGGID_OFFSET_FLAGCHECK + 11
# all possible issue IDs (used for initialization of issue list)
ALL_MSGID_FLAGCHECK = (
    MSGID_FLAGCHECK_ILLEGAL, MSGID_FLAGCHECK_MISSING_100,
    MSGID_FLAGCHECK_VALID_VALUE_MISSIG_FLAG,
    MSGID_FLAGCHECK_MISSING_VALUE_FLAG_100,
    MSGID_FLAGCHECK_MISSING_VALUE_NO_FLAG,
    MSGID_FLAGCHECK_PRECIP_FLAG_782_783_784_DROPPED,
    MSGID_FLAGCHECK_PRECIP_FLAG_782_783_784_ERROR,
    MSGID_FLAGCHECK_PRECIP_FLAG_890_DROPPED,
    MSGID_FLAGCHECK_PRECIP_FLAG_890_SET_ZERO,
    MSGID_FLAGCHECK_PRECIP_FLAG_890_ERROR,
    MSGID_FLAGCHECK_AUX_VALID_VALUE_MISSING_FLAG)

# value check issues (boundaries and spikes), used in
# ebas.domain.basic_domain_logic.value_check
MSGID_OFFSET_VALUECHECK = 100
MSGID_VALUECHECK_BNDCK_LOWER = MSGID_OFFSET_VALUECHECK + 1
MSGID_VALUECHECK_BNDCK_UPPER = MSGID_OFFSET_VALUECHECK + 2
MSGID_VALUECHECK_SPKCK_LOWER = MSGID_OFFSET_VALUECHECK + 3
MSGID_VALUECHECK_SPKCK_UPPER = MSGID_OFFSET_VALUECHECK + 4
MSGID_VALUECHECK_SPKCK_STD_LOWER = MSGID_OFFSET_VALUECHECK + 5
MSGID_VALUECHECK_SPKCK_STD_UPPER = MSGID_OFFSET_VALUECHECK + 6
# all possible issue IDs (used for initialization of issue list)
ALL_MSGID_VALUECHECK = (
    MSGID_VALUECHECK_BNDCK_LOWER, MSGID_VALUECHECK_BNDCK_UPPER,
    MSGID_VALUECHECK_SPKCK_LOWER, MSGID_VALUECHECK_SPKCK_UPPER,
    MSGID_VALUECHECK_SPKCK_STD_LOWER, MSGID_VALUECHECK_SPKCK_STD_UPPER)

# boundary and spike check selection issues, used in
# ebas.io.file.basefile.ceck_data and ebas.domain.entities.ts
MSGID_OFFSET_BCSEL = 100
MSGID_BCSEL_NO_PROJ = MSGID_OFFSET_BCSEL + 1
MSGID_BCSEL_DUPLICATES = MSGID_OFFSET_BCSEL + 2
ALL_MSGID_BCSEL = (MSGID_BCSEL_NO_PROJ, MSGID_BCSEL_DUPLICATES)


class EbasDomainIssueList(ControlledDictList):
    """
    List of issues to be reported to a caller. Multiple errors and warnings are
    possible. The caller can decide whhat to do with the issues (usually
    condense and log them).
    """
    def __init__(self, *args, **kwargs):
        """
        Set up the issue list.
        Parameters:
            keys    key names used for the elements of teach issue
        Returns:
            none
        """
        self.allowed = {}
        if 'keys' not in kwargs:
            # default keys for issue list
            kwargs['keys'] = ('severity', 'msg_id', 'row', 'message')
        if 'msg_id' not in kwargs or not kwargs['msg_id']:
            kwargs['msg_id'] = ALL_MSGID_FLAGCHECK + ALL_MSGID_VALUECHECK + \
                ALL_MSGID_BCSEL
        for key in kwargs['keys']:
            if key not in kwargs and key == 'severity':
                # default severities:
                kwargs[key] = [ISSUE_SEVERITY_ERROR,
                               ISSUE_SEVERITY_WARNING]
        super(EbasDomainIssueList, self).__init__(*args, **kwargs)



"""
$Id: base.py 2088 2018-06-29 13:46:38Z pe $

Base class for EBAS Masterdata Classes
"""

class EbasMasterBase(object):
    """
    Domain Class for statistics codes masterdata.
    Objects of this class do not represent entities, this class provides class
    methods for handling statistics codes and checking them against master data.
    Statistics codes master data are retrieved from database or from offline
    storage when no database access is possible.
    """

    def __getitem__(self, key):
        """
        Allows dictionary like access to metadata
        """
        return self.__class__.META[key]

    def __iter__(self):
        """
        Iterator for statistics codes masterdata
        """
        for key in sorted(self.__class__.META.keys()):
            yield self.__class__.META[key]




"""
$Id: commandline.py 2454 2020-04-24 20:22:51Z pe $

wrapper for nilu command line programs.
"""
import sys
import errno
import logging
import argparse
import getpass
import textwrap
import netrc
import cProfile, pstats
import datetime
import re
import os.path
from nilutility.argparse_helper import NiluArgumentParser, \
    ParseStrings, ParseIntegers, \
    parse_time_instant, parse_time_interval
from nilutility import logging_helper
from nilutility.encoding import fix_io_encoding
from nilutility.datetime_helper import DatetimeInterval

# ConfigParser has been renamed to configparser in python3
# when dropping python to -> import configparser
from six.moves import configparser


class NiluCommandline(object):  # pylint: disable-msg=R0902
    # R0902: Too many instance attributes
    """
    class for handling Nilu commandline arguments.
    """
    def __init__(self, func, custom_args=None,  # pylint: disable-msg=R0913
                 private_args=None, help_description=None,
                 version='#no-version-number',
                 version_date=datetime.datetime.now(),
                 logging_params = None,
                 progname=None, config_main_section='GENERAL',
                 def_args=None):
        # R0913: Too many arguments
        self.custom_args = custom_args or []
        self.private_args = private_args
        self.help_description = help_description
        self.version = version
        self.version_date = version_date
        self.func = func
        self.unparsed = None
        if progname:
            self.progname = progname
        else:
            self.progname = func.__name__  # func can be a class or a function
        self.config_main_section = config_main_section
        self.startuptime = datetime.datetime.now()
        self.logbasename = self.progname + \
                           self.startuptime.strftime('_%Y-%m-%d_%H%M%S')
        self.logging_params = logging_params
        self.logger = None
        self.args = None
        self.def_args = def_args
        self.set_defaults()
        self.versiontext=textwrap.dedent('''\n                                   
            %(prog)s {0}                                                    
            Copyright (C) 2012-{1} NILU - Norwegian Institut for Air Reserch     
            '''.format(self.version, self.version_date.year))
        self.logger = logging.getLogger('NiluCmdLine')

    def run(self):
        self.unparsed = sys.argv[1:]
        self.preparse()
        self.parse_args()
        self.logger.debug("starting {}, version {}".format(self.progname,
                                                           self.version))
        self.wrap_func()

    def preparse(self):
        if 'CONFIG' in self.custom_args:
            self.preparse_config()
        if 'LOGGING' in self.custom_args:
            self.preparse_logging()

    def set_defaults(self):
        """
        Set defaults.

        Parameters:
            None
        Returns:
           None
        """
        default = {}   # the global nilutility default
        if 'CONFIG' in self.custom_args:
            default['cfgfile'] = '{}.cfg'.format(self.progname)
        if 'LOGGING' in self.custom_args:
            default['loglevfile'] = 'silent'
            default['loglevconsole'] = 'info'
            default['logfile'] = os.path.join(
                '.', self.logbasename) + '.log'
        if self.def_args:
            # update with default from caller (if set)
            default.update(self.def_args)
        self.def_args = default

    def _preparse_config_prepare(self):
        """
        Prepare the configuration file (find the file and open it).
        This logig is rather general and can easyly be reused by derived
        classes.
        Parameters:
            None
        Returns:
            config parser object
            None if any problem with the config file
        """
        preparser = NiluArgumentParser(add_help=False)
        preparser.add_config_arguments(cfgfile=True)
        preparser.set_known_defaults(**self.def_args)
        self.args, self.unparsed = preparser.parse_known_args(
            args=self.unparsed, namespace=self.args)
        # expand ~ to users directories ~ home, ~user other user's home:
        self.args.cfgfile = os.path.expanduser(self.args.cfgfile)
        # read config
        try:
            fil = open(self.args.cfgfile, "r")
            if (os.sep not in self.args.cfgfile and
                os.path.realpath(os.path.expanduser(
                    '~/'+self.args.cfgfile)) !=
                    os.path.realpath((self.args.cfgfile))):
                sys.stderr.write('WARNING: using local config file {}\n'.format(
                    self.args.cfgfile))
            self.args.cfgfile = fil
        except IOError as expt:
            if expt.errno == 2:
                if os.sep in self.args.cfgfile:
                    sys.stderr.write(
                        "no config file ({}) found; using defaults".format(
                            self.args.cfgfile) + '\n')
                    self.args.cfgfile = None
                    return None
                else:
                    # relative filename: when current dir failed, try
                    # users home dir, finally program dir (thus, a default
                    # config can be provided with the installation)

                    # try program's directory:
                    localname = os.path.expanduser('~/' + self.args.cfgfile)
                    try:
                        fil = open(localname, 'r')
                        # no warning when using home dir config file
                        self.args.cfgfile = fil
                    except IOError as expt:
                        if expt.errno == 2:
                            # try program's directory:
                            localname = os.path.join(os.path.dirname(
                                os.path.realpath(sys.argv[0])),
                                self.args.cfgfile)
                            try:
                                fil = open(localname, 'r')
                                sys.stderr.write(
                                    'WARNING: using program global config file '
                                    '{}\n'.format(localname))
                                self.args.cfgfile = fil
                            except IOError as expt:
                                if expt.errno == 2:
                                    sys.stderr.write(
                                        "no config file (./{0}, ~/{0}, {1}) "
                                        "found; using defaults".format(
                                            self.args.cfgfile, localname) + \
                                        '\n')
                                    self.args.cfgfile = None
                                    return None
                                else:
                                    sys.stderr.write(
                                        "config file ({}) error: {}; "
                                        "using defaults\n".format(
                                            localname, str(expt)))
                                    self.args.cfgfile = None
                                    return None
                        else:
                            sys.stderr.write(
                                "config file ({}) error: {}; using defaults\n"\
                                    .format(localname, str(expt)))
                            self.args.cfgfile = None
                            return None

            else:
                sys.stderr.write("config file error: " + str(expt) + \
                                 '; using defaults\n')
                return None
        config = configparser.ConfigParser()
        config.optionxform = str  # make config options case sensitive!
        try:
            config.readfp(fil)
        except AttributeError:
            # from python 3.12, ConfigParser has no readfp
            config.read_file(fil)
        # reset the file, in case the app wants to use the config file:
        fil.seek(0)
        return config

    def preparse_config(self):
        """
        Partially parse commandline arguments relevant config file location.
        Then read the config file.
        preparser: process only --cfgfile
        Needs to be pre-parsed in order to use defaults in the real
        parser.
        """
        # prepare the config file parser:
        config = self._preparse_config_prepare()
        if config is None:
            return

        try:
            self.def_args.update(dict(config.items(self.config_main_section)))
        except configparser.Error as expt:
            sys.stderr.write('config file {}; error: {}\n'.format(
                self.args.cfgfile, str(expt)) + '\n')
        if self.config_main_section == self.progname:
            return

        try:
            self.def_args.update(dict(config.items(self.progname)))
        except configparser.NoSectionError:
            pass
        except configparser.Error as expt:
            sys.stderr.write('config file {}; error: {}\n'.format(
                self.args.cfgfile, str(expt)) + '\n')

    def add_custom_args(self, parser):                                           
        """                                                                      
        Add arguments for specific customized goals.

        Parameters:
            parser    parser object
        Returns:
            None
        """
        parser.add_argument('--version', '-V ', action='version', \
            version=self.versiontext,
            help='display version information and exit')
        if 'CONFIG' in self.custom_args:
            parser.add_config_arguments(cfgfile=True)
        if 'LOGGING' in self.custom_args:
            parser.add_logging_arguments(logfile=True, loglevconsole=True,
                                         loglevfile=True, profile=True)

    def preparse_logging(self):
        """
        Partially parse commandline arguments relevant for logging, then init
        logging according to setings.
        """
        preparser = NiluArgumentParser(add_help=False)
        preparser.add_logging_arguments(logfile=True, loglevconsole=True,
                                        loglevfile=True, profile=True)
        preparser.set_known_defaults(**self.def_args)
        self.args, self.unparsed = preparser.parse_known_args(
            args=self.unparsed, namespace=self.args)

        # init logging
        console={'level': self.args.loglevconsole,
                 'format': '%(levelname)-8s: %(message)s'}
        if self.logging_params and 'console' in self.logging_params:
            console.update(self.logging_params['console'])
        logfile={'file': self.args.logfile,
                 'createdirs': True,
                 'level': self.args.loglevfile}
        if self.logging_params and 'file' in self.logging_params:
            logfile.update(self.logging_params['file'])
        try:
            logging_helper.init_logging(console=console, logfile=[logfile])
        except IOError as expt:
            sys.stderr.write("log file error: {}; exiting\n".format(str(expt)))
            sys.exit(expt.errno)  # IOError errcodes are all positive integers

    def parse_args(self):
        """
        Main argument parsing.

        Parameters:
            None
        Returns:
           None
        """
        parser = NiluArgumentParser(add_help=True,
                                    description=self.help_description)

        # custom parsers (parser groups used for the same purpose in many
        # programs)
        self.add_custom_args(parser)
        # callback function for argumenst used only by the calling program
        if self.private_args:
            self.private_args(parser, self)

        parser.set_known_defaults(**self.def_args)
        self.args = parser.parse_args(args=self.unparsed,
                                      namespace=self.args)

    def wrap_func(self):
        """
        Wrapper for nilu commandline programs. Catches Ctrl-C and broken pipe.
        Catches sys.exit from below and finishes profiling in those cases before
        re-raising.
        Adds profiling if commandline option --profile is given

        Parameters:
            func      function to be called (main execution for the program)
            args      args to be used for function call
            kwargs    kwargs to be used for function call

        Returns:
            None

        Exits:
            with exit code in case of user interrupt or broken pipe
        """

        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument('--profile', action='store_true')
        arg, _ = parser.parse_known_args()
        sysexit = None
        if arg.profile:
            prof = cProfile.Profile()
            prof.enable()
        try:
            self.func(self)
        except KeyboardInterrupt:
            logger = logging.getLogger('NiluMain')
            logger.fatal('terminated on user request')
            sys.exit(1)
        except IOError as expt:
            if expt.errno == errno.EPIPE:
                logger = logging.getLogger('NiluMain')
                logger.fatal('broken pipe, exit')
                sys.exit(expt.errno)
            else:
                raise
        except SystemExit as expt:
            logger = logging.getLogger('NiluMain')
            logger.debug("%s terminated with exit code %s", self.progname,
                         str(expt.code))
            sysexit = expt
        if arg.profile:
            prof.disable()
            profile_file = self.logbasename + '.profstat'
            self.logger.info('writing profiling statistics to file %s',
                             profile_file)
            self.logger.info('writing profiling top 10% sorted by time to '
                             'stderr')
            prof.dump_stats(profile_file)
            sortby = 'time'
            pstat = pstats.Stats(prof, stream=sys.stderr).sort_stats(sortby)
            pstat.print_stats(.1)
        if sysexit:
            raise sysexit # pylint: disable-msg=E0702
            # E0702: Raising NoneType
            #    this is conditional. Pylint bug.

fix_io_encoding('UTF-8')



"""
EANET filter file format
   base class

$Id: base.py 2762 2021-11-30 23:42:49Z pe $
"""

import logging
import re
import datetime
from collections import OrderedDict

def colnum_string(col):
    """
    Get the column strring for a spreadsheet adddress.
    Parameters:
        col    column number (0 based)
    Returns:
        Spreadsheet column address string A, B, ..., Z, AB, ...
    """
    col += 1
    string = ""
    while col > 0:
        col, remainder = divmod(col - 1, 26)
        string = chr(65 + remainder) + string
    return string

def column_string_to_num(string):
    """

    """
    col = ord(string[-1]) - 65
    if string[:-1]:
        return 26 * (column_string_to_num(string[:-1]) + 1) + col
    return col

def get_address(row, col):
    """
    Translate 0 based row and col to spreadsheet adddress, e.g D25.
    Parameters:
        row    row number (0 based)
        col    column number (0 based)
    Returns:
        Spreadsheet address (e.g. "A25")
    """
    return colnum_string(col) + str(row+1)

def from_address(address):
    """
    Converts spreadsheet notation (e.g. A1) to row and col number.
    Parameters:
        address    spreadsheet notation string, e.g. A1
    Returns:
        tuple, row and col number, both 0 based (row, col)
    """
    reg = re.match(r'^([A-Z]+)([0-9]+)$', address)
    if not re:
        raise ValueError('{} is not in spreadsheet notation'.format(address))
    return(int(reg.group(2))-1, column_string_to_num(reg.group(1)))


class EanetError(Exception):
    """
    Base error class for this module.
    """
    pass


class EanetFilterVariable(object):  # pylint: disable=R0903
    # R0903: Too few public methods
    #     --> this is a base class
    """
    One variable in a filer file.
    """

    def __init__(self, name, unit, samples, flags):
        """
        Initialize the object.
        """
        self.logger = logging.getLogger("EanetFilterVar")
        self.name = name
        self.unit = unit
        self.samples = samples
        self.flags = flags


class EanetPrecipVariable(object):  # pylint: disable=R0903
    # R0903: Too few public methods
    #     --> this is a base class
    """
    One variable in a filer file.
    """

    def __init__(self, name, unit, samples, flags):
        """
        Initialize the object.
        """
        self.logger = logging.getLogger("EanetFilterVar")
        self.name = name
        self.unit = unit
        self.samples = samples
        self.flags = flags

class EanetBase(object):
    """
    Base class for all EANET files.
    """
    def __init__(self):
        """
        Initialize the object.
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.errors = 0
        self.warnings = 0

    def message(self, severity, errtxt, *args, **kwargs):
        """
        Log a message with reference to row and column
        Parameters:
            severity  logging severity
            errtxt    error text
            *args     optional additional arguments passed to logger.log
            row       row number (0 based)
            col       col number (0 based)
        Note: using kwargs instead og row=None, col=None because this form is
        not allowed in py2.
        Returns:
            None
        """
        # some manual parameter handling (see above, restrictions in py2)
        illegal = [x for x in kwargs if x not in ('row', 'col')]
        if illegal:
            raise TypeError('unexpected keyword arguments {}'.format(
                ', '.join(illegal)))
        row = kwargs.get("row", None)
        col = kwargs.get("col", None)

        spec = ''
        if row is not None:
            spec += "Row {}".format(row+1)
            if col is not None:
                spec += ', col {} [{}]: '.format(col+1, get_address(row, col))
        elif col is not None:
            spec += "Col %s [%s]: ".format(col+1, colnum_string(col))
        if 'row' in kwargs:
            del kwargs['row']
        if 'col' in kwargs:
            del kwargs['col']
        self.logger.log(severity, "%s"+errtxt, spec, *args, **kwargs)

    def error(self, errtxt, *args, **kwargs):
        """
        Log an error with reference to row and column
        Parameters:
            errtxt    error text
            *args     optional additional arguments passed to logger.log
            row       row number (0 based)
            col       col number (0 based)
        Returns:
            None
        """
        self.errors += 1
        self.message(logging.ERROR, errtxt, *args, **kwargs)

    def warning(self, errtxt, *args, **kwargs):
        """
        Log a warning with reference to row and column
        Parameters:
            errtxt    error text
            *args     optional additional arguments passed to logger.log
            row       row number (0 based)
            col       col number (0 based)
        Returns:
            None
        """
        self.warnings += 1
        self.message(logging.WARNING, errtxt, *args, **kwargs)

    def info(self, errtxt, *args, **kwargs):
        """
        Log an info with reference to row and column
        Parameters:
            errtxt    error text
            *args     optional additional arguments passed to logger.log
            row       row number (0 based)
            col       col number (0 based)
        Returns:
            None
        """
        self.message(logging.INFO, errtxt, *args, **kwargs)

    def _parse_datetime(self, row, col):
        """
        Parse date/time object form row,col (date) and row,col+1 (time)
        Parameters:
            row   row in rawdata
            col   column in rawdata (date), col+1 is time
        Expected date format: YYYY/MM/DD
        Expected time format: HH:MI
        """
        reg = re.match(
            r'^(\d+)\/(\d+)\/(\d+)$', self._rawdata[row][col])
        if not reg:
            self.error(
                "Date format error '%s'", self._rawdata[row][col],
                row=row, col=col)
            return None
        else:
            year, mon, day = (int(x) for x in reg.groups())

        if not self._rawdata[row][col+1]:
            # interpret empty as 00:00
            hour = 0
            minute = 0
        else:
            reg = re.match(
                r'^(\d+):(\d+)$', self._rawdata[row][col+1])
            if not reg:
                self.error(
                    "Time format error '%s'", self._rawdata[row][col],
                    row=row, col=col+1)
                return None
            else:
                hour, minute = (int(x) for x in reg.groups())

        try:
            return datetime.datetime(year, mon, day, hour, minute)
        except ValueError as excpt:
            self.error(
                "Datet time error: '%s'", excpt,
                row=row, col=col+1)
            return None


class EanetFilterBase(EanetBase):  # pylint: disable=R0902, R0903
    # R0902: Too many instance attributes
    # R0903: Too few public methods
    #     --> this is a base class
    """
    Base class for EANET filter files.
    """

    def __init__(self):
        """
        Initialize the object.
        """
        super(EanetFilterBase, self).__init__()
        self.file_name = None
        self.country_name = None
        self.site_code = None
        self.site_name = None
        self.sample_times = []
        self.variables = []  # list of EanetFilterVariable
        self.notes = None  # None if no notes col in file, [] if notes
        self._hasnotes = None
        self._lastcol = None

class EanetPrecipForm(object):
    """
    Object representing a form in the document.
    """

    def __init__(self, parent, index_in_file, name, layout):
        """
        Set up the object.
        Parameters:
            parent         The parent file object
            index_in_file  Index of this object in the parent
            name           Name of this form
            layout         The layout of the form (part from file's LAYOUT)
        Returns:
            None
        """
        self.file = parent
        self.index_in_file = index_in_file
        self.name = name
        self.layout = layout
        self.variables = []

    def var_by_name(self, name):
        """
        Get a variable by name.
        (Variable name are unique withing a form (this is checked at class
        creation).
        Parameters:
            name    variable name from the LAYOUT
        Returns:
            EantePrecipVariable object
        Raises:
            KeyError
        """
        for var in self.variables:
            if var.name == name:
                return var
        raise KeyError("No variable named " + name)


class EanetPrecipBase(EanetBase):  # pylint: disable=R0902, R0903
    # R0902: Too many instance attributes
    # R0903: Too few public methods
    #     --> this is a base class
    """
    Base class for EANET filter files.
    """

    LAYOUT = OrderedDict([
        (
            u'Form (Wet A)  No. 1   Results of wet deposition analysis '
            u'(Anion)',
            {
                'header': "B2",
                'start': "B11",
                'sample_number': "Sample No.",
                'sample_times': True,
                'vars': [  # --> OrderedDict
                    # will be changed to OrderedDict after checking for
                    # uniqueness of keys!
                    (u"SO42-",
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'NO3-',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'Cl-',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'HCO3-',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'F-',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'Br-',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'NO2-',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'PO43-',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 1>',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 2>',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'Date of analysis',
                     {
                         'flags': 0,
                         'merge': 'concat',
                     }),
                    (u'Note',
                     {
                         'flags': 0,
                         'merge': 'concat',
                     }),
                ],
            }
        ),

        (
            u'Form (Wet A)  No. 2   Results of wet deposition analysis '
            u'(Cation)',
            {
                'header': "AX2",
                'start': "AX11",
                'sample_number': "Sample No.",
                'sample_times': True,
                'vars': [  # --> OrderedDict
                    # will be changed to OrderedDict after checking for
                    # uniqueness of keys!
                    (u'NH4+',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'Na+',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'K+',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'Ca2+',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'Mg2+',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 1>',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 2>',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'Date of analysis',
                     {
                         'flags': 0,
                         'merge': 'concat',
                     }),
                    (u'Note',
                     {
                         'flags': 0,
                         'merge': 'concat',
                     }),
                ],
            }
        ),

        (
            u'Form (Wet A)  No. 3   Results of wet deposition analysis '
            u'(EC, pH, R1, R2, Precipitation)',
            {
                'header': "CH2",
                'start': "CH11",
                'sample_number': "Sample No.",
                'sample_times': True,
                'vars': [  # --> OrderedDict
                    # will be changed to OrderedDict after checking for
                    # uniqueness of keys!
                    (u'EC',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'pH',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'R1',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'R2',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'Amount of sample',
                     {
                         'flags': 3,
                         'merge': 'sum',
                     }),
                    (u'Amount of precipitation',
                     {
                         'flags': 0,
                         'merge': 'sum',
                         'is_amount': True,
                     }),
                    (u'Method Code',
                     {
                         'flags': 3,
                         'merge': 'first',
                     }),
                    (u'Date of analysis',
                     {
                         'flags': 0,
                         'merge': 'concat',
                     }),
                    (u'Note',
                     {
                         'flags': 0,
                         'merge': 'concat',
                     }),
                ],
            }
        ),

        (
            u'Form (Wet A)  No. 4   Results of wet deposition analysis '
            u'(Organic ions)',
            {
                'header': "DO2",
                'start': "DO11",
                'sample_number': "Sample No.",
                'sample_times': True,
                'vars': [  # --> OrderedDict
                    # will be changed to OrderedDict after checking for
                    # uniqueness of keys!
                    (u'HCOO-',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'CH3COO-',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'(COO- )2',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 1>',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 2>',
                     {
                         'flags': 3,
                         'merge': 'mean',
                     }),
                    (u'Date of analysis',
                     {
                         'flags': 0,
                         'merge': 'concat',
                     }),
                    (u'Note',
                     {
                         'flags': 0,
                         'merge': 'concat',
                     }),
                ],
            }
        ),

        (
            u'Summary 1',
            {
                'header': "EQ2",
                'start': "EQ11",
                'sample_number': "Sample\n No.",
                'sample_times': False,
                'vars': [  # --> OrderedDict
                    # will be changed to OrderedDict after checking for
                    # uniqueness of keys!
                    (u'nss-SO42-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'nss-Ca2+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'H+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'HCOO-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'CH3COO-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'   COOH\n   COO-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'(COO- )2',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 1>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 2>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                ],
            }
        ),

        (
            u'Summary 2',
            {
                'header': "FB2",
                'start': "FB11",
                'sample_number': "Sample\n No.",
                'sample_times': False,
                'vars': [  # --> OrderedDict
                    # will be changed to OrderedDict after checking for
                    # uniqueness of keys!
                    (u'SO42-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'NO3-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Cl-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'HCO3-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'F-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Br-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'NO2-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'PO43-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 1>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 2>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'HCOO-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'CH3COO-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'(COO- )2',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 3>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 4>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Anion',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'NH4+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Na+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'K+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Ca2+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Mg2+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 5>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 6>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'H+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Cation',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'C+A',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'nss-SO42-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'nss-Ca2+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'EC x ppt.',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                ],
            }
        ),

        (
            u'Summary 3',
            {
                'header': "GG2",
                'start': "GG11",
                'sample_number': "Sample\n No.",
                'sample_times': False,
                'vars': [  # --> OrderedDict
                    # will be changed to OrderedDict after checking for
                    # uniqueness of keys!
                    (u'SO42-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'NO3-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Cl-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'HCO3-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'F-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Br-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'NO2-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'PO43-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 1>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 2>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'HCOO-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'CH3COO-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'(COO- )2',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 3>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 4>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Anion',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'NH4+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Na+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'K+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Ca2+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Mg2+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 5>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 6>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'H+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Cation',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'C+A',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Req. R1',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'R1',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                ],
            }
        ),

        (
            u'Summary 4',
            {
                'header': "HK2",
                'start': "HK11",
                'sample_number': "Sample No.",
                'sample_times': False,
                'vars': [  # --> OrderedDict
                    # will be changed to OrderedDict after checking for
                    # uniqueness of keys!
                    (u'SO42-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'NO3-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Cl-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'HCO3-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'F-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Br-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'NO2-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'PO43-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 1>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 2>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'HCOO-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'CH3COO-',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'(COO- )2',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 3>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 4>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Anion',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'NH4+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Na+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'K+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Ca2+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Mg2+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 5>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'<EMPTY 6>',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'H+',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Cation',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'ECcal',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'ECmes',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Req. R2',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'R2',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Amount of ppt. (cal)',
                     {
                         'flags': 0,
                         'merge': 'sum',
                     }),
                    (u'Amount of ppt.',
                     {
                         'flags': 0,
                         'merge': 'sum',
                     }),
                    (u'%CE  (M/R)',
                     {
                         'flags': 0,
                         'merge': 'mean',
                     }),
                    (u'Sampling period',
                     {
                         'flags': 0,
                         'merge': 'sum',
                     }),
                ],
            }
        ),
    ])
    for _, form in LAYOUT.items():
        # Make sure var names are unique!
        names = [var[0] for var in form['vars']]
        if len(names) != len(set(names)):
            raise RuntimeError(
                "EanetPrecipBase.LAYOUT misconfigured (non unique var name)")
        # Then make an OrderedDict.
        form['vars'] = OrderedDict(form['vars'])


    def __init__(self):
        """
        Initialize the object.
        """
        super(EanetPrecipBase, self).__init__()
        self.file_name = None
        self.site_code = None
        self.forms = []  # different forms of the sheet

        self.country_name = None
        self.site_name = None
        self.laboratory = None
        self.reporter = None
        self.funnel_diameter = None

        self.sample_numbers = []
        self.sample_times = []



"""
$Id: dl.py 2088 2018-06-29 13:46:38Z pe $

EBAS Masterdata Class for data level masterdata

This module implements the class EbasMasterDL.
"""

from .offline_masterdata import DLOfflineMasterData
from .base import EbasMasterBase

class EbasMasterDL(EbasMasterBase, DLOfflineMasterData):
    """
    Domain Class for data level masterdata
    Objects of this class do not represent entities, this class provides class
    methods for handling data levels and checking them against master data.
    Data level master data are retrieved from database or from offline storage
    when no database access is possible.
    """

    # read offline masterdata for DL (data levels)
    # Those are fallback values, will be read from database as soon as possible.
    DLOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read co masterdata if dbh is provided.
        """
        DLOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)



"""
$Id: am.py 2088 2018-06-29 13:46:38Z pe $

EBAS Masterdata Class for analyticalinstrument manufacturer/model masterdata

This module implements the class EbasMasterAM.

History:
V.1.0.0  2016-10-17  pe  initial version

"""

from .offline_masterdata import AMOfflineMasterData
from .base import EbasMasterBase

class EbasMasterAM(EbasMasterBase, AMOfflineMasterData):
    """
    Domain Class for analyticalinstrument manufacturer/model masterdata
    Objects of this class do not represent entities, this class provides class
    methods for handling components and checking them against master data.
    Components master data are retrieved from database or from offline storage
    when no database access is possible.
    """

    # read offline masterdata for AM (analyticalinstrument manufacturer/model)
    # Those are fallback values, will be read from database as soon as possible.
    AMOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read am masterdata if dbh is provided.
        """
        AMOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)

    def list(self, at_type=None, am_manufacturer=None, am_model=None):
        """
        Generates a list of matching masterdata.
        Parameters:
            at_type           analyticalinstrument type
            am_manufacturer   instrument manufacturer
            am_model          instrument model name
        Returns:
            generator of AM dictionaries with matching criteria
        """
        for am_ in self.__class__.META.values:
            if (at_type is None or am_['AT_TYPE'] == at_type) and \
               (am_manufacturer is None or \
                am_['AM_MANUFACTURER'] == am_manufacturer) and \
               (am_model is None or am_['AM_MODEL'] == am_model):
                yield am_
